import json
from hashlib import md5
from unittest.mock import Mock, patch

import pytest

from coltrane.management.commands.build import Command
from coltrane.manifest import Manifest


@pytest.fixture
def build_command():
    cmd = Command()
    cmd.is_force = False

    return cmd


def _reset_settings(settings, tmp_path):
    settings.BASE_DIR = tmp_path
    settings.STATIC_ROOT = tmp_path / "output" / "static"


def create_markdown_file(tmp_path):
    (tmp_path / "content").mkdir()
    markdown_file = tmp_path / "content" / "test-1.md"
    markdown_file.write_text("# test 1")

    return markdown_file


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_force_false(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    build_command.handle(force=False)

    assert build_command.is_force is False


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_force_true(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    build_command.handle(force=True)

    assert build_command.is_force is True


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._load_manifest", spec=Manifest)
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_static_files_changed_is_force(_load_manifest, settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    _load_manifest.return_value.static_files_manifest_changed = True

    build_command.handle(force=False)

    assert build_command.is_force is True


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_create(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    create_markdown_file(tmp_path)

    build_command.handle(force=False)

    assert build_command.output_result_counts.create_count == 1
    assert build_command.output_result_counts.update_count == 0
    assert build_command.output_result_counts.skip_count == 0


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_update(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    create_markdown_file(tmp_path)

    # Fake a previous run with the invalid metadata
    (tmp_path / "output.json").write_text(json.dumps({"test-1.md": {"mtime": -1, "md5": "not-a-hash"}}))

    build_command.handle(force=False)

    assert build_command.output_result_counts.create_count == 0
    assert build_command.output_result_counts.update_count == 1
    assert build_command.output_result_counts.skip_count == 0


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_skip_because_mtime(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    markdown_file = create_markdown_file(tmp_path)

    # Create output.json
    (tmp_path / "output.json").write_text("{}")

    # Fake a previous run with the correct md5
    md5_hash = md5(markdown_file.read_bytes()).hexdigest()  # noqa: S324
    (tmp_path / "output.json").write_text(json.dumps({"test-1.md": {"mtime": -1, "md5": md5_hash}}))

    build_command.handle(force=False)

    assert build_command.output_result_counts.create_count == 0
    assert build_command.output_result_counts.update_count == 0
    assert build_command.output_result_counts.skip_count == 1


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_skip_because_md5(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    markdown_file = create_markdown_file(tmp_path)

    # Create output.json
    (tmp_path / "output.json").write_text("{}")

    # Fake a previous run with the correct mtime
    mtime = markdown_file.stat().st_mtime
    (tmp_path / "output.json").write_text(json.dumps({"test-1.md": {"mtime": mtime, "md5": "not-a-hash"}}))

    build_command.handle(force=False)

    assert build_command.output_result_counts.create_count == 0
    assert build_command.output_result_counts.update_count == 0
    assert build_command.output_result_counts.skip_count == 1


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_threads(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    build_command.handle(threads=3)

    assert build_command.threads_count == 3


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_invalid_threads_count(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    build_command.handle(threads="asdf")

    assert build_command.threads_count == 2  # default number of threads


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.cpu_count")
def test_handle_cpu_count_exception(cpu_count, settings, tmp_path, build_command):
    cpu_count.side_effect = Exception()
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    build_command.handle()

    assert build_command.threads_count == 2  # default number of threads


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_template_error_exit_with_1(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Force debug to be true to surface template error
    settings.DEBUG = True

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("{{sadf}}")

    with pytest.raises(SystemExit) as err:
        build_command.handle()

    assert err.type == SystemExit
    assert err.value.code == 1


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_template_error_formatted(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Force debug to be true to surface template error
    settings.DEBUG = True

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("{{sadf}}")

    with pytest.raises(SystemExit):
        build_command.handle()

    assert len(build_command.errors) == 1
    assert "'sadf' does not exist in template context. Available top level variables:" in build_command.errors[0]


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
def test_handle_ignore_template_error(settings, tmp_path, build_command):
    _reset_settings(settings, tmp_path)

    # Force debug to be true to surface template error
    settings.DEBUG = True

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("{{sadf}}")

    # Ignore the template error so exit isn't called
    build_command.handle(ignore=True)


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.Command._generate_sitemap")
def test_handle_generate_sitemap(
    _generate_sitemap,
    settings,
    tmp_path,
    build_command,
):
    _reset_settings(settings, tmp_path)

    create_markdown_file(tmp_path)

    build_command.handle(force=False)

    _generate_sitemap.assert_called_once()


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.Command._generate_rss")
def test_handle_generate_rss(
    _generate_rss,
    settings,
    tmp_path,
    build_command,
):
    _reset_settings(settings, tmp_path)

    create_markdown_file(tmp_path)

    build_command.handle(force=False)

    _generate_rss.assert_called_once()



from copy import deepcopy
from os import environ
from pathlib import Path
from unittest.mock import ANY, Mock, patch

from django.template.library import InvalidTemplateLibrary

from coltrane import (
    DEFAULT_CACHES_SETTINGS,
    DEFAULT_COLTRANE_SETTINGS,
    DEFAULT_INSTALLED_APPS,
    DEFAULT_MIDDLEWARE,
    # _get_default_template_settings,
    initialize,
)
from coltrane.config.settings import get_config

base_dir = Path(".")
config = get_config()

DEFAULT_SETTINGS = {
    "BASE_DIR": base_dir,
    "ROOT_URLCONF": "coltrane.urls",
    "DEBUG": True,
    "SECRET_KEY": ANY,
    "INSTALLED_APPS": deepcopy(DEFAULT_INSTALLED_APPS),
    "CACHES": deepcopy(DEFAULT_CACHES_SETTINGS),
    "MIDDLEWARE": deepcopy(DEFAULT_MIDDLEWARE),
    "TEMPLATES": deepcopy(config.get_templates_settings()),
    "INTERNAL_IPS": [],
    "ALLOWED_HOSTS": [],
    "STATIC_ROOT": ANY,
    "STATIC_URL": ANY,
    "STATICFILES_DIRS": ANY,
    "TIME_ZONE": ANY,
    "LOGGING": ANY,
    "COLTRANE": deepcopy(DEFAULT_COLTRANE_SETTINGS),
    "SETTINGS_MODULE": "coltrane",
    "ENV": ANY,
}


def _get_settings():
    settings = deepcopy(DEFAULT_SETTINGS)

    settings["INSTALLED_APPS"].append("django_browser_reload")
    settings["MIDDLEWARE"].append("django_browser_reload.middleware.BrowserReloadMiddleware")

    return settings


def _get_settings_with_debug_false():
    settings = deepcopy(DEFAULT_SETTINGS)
    settings["DEBUG"] = False

    return settings


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_no_base_dir(_configure_settings, *args):
    initialize()

    expected = _get_settings()

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_with_base_dir(_configure_settings, *args):
    initialize(BASE_DIR=Path("test"))

    expected = _get_settings()
    expected["BASE_DIR"] = Path("test")
    expected["TEMPLATES"][0]["DIRS"] = [Path("test") / "templates"]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_with_base_dir_as_string(_configure_settings, *args):
    initialize(BASE_DIR="test")

    expected = _get_settings()
    expected["BASE_DIR"] = Path("test")
    expected["TEMPLATES"][0]["DIRS"] = [Path("test") / "templates"]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(return_value="fake.templatetag"),
)
@patch("coltrane._configure_settings")
def test_with_template_tags(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY
    expected["TEMPLATES"][0]["OPTIONS"]["builtins"].append("fake.templatetag")

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(side_effect=InvalidTemplateLibrary),
)
@patch("coltrane._configure_settings")
def test_with_invalid_template_tag(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(return_value="fake.templatetag"),
)
@patch("coltrane._configure_settings")
def test_with_template_tags_in_directory_with_py_extension(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    (tmp_path / "templatetags" / "sample.py").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample.py" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY
    expected["TEMPLATES"][0]["OPTIONS"]["builtins"].append("fake.templatetag")

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_debug_setting(_configure_settings):
    initialize(DEBUG=False)

    expected = _get_settings_with_debug_false()
    expected["TEMPLATES"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch.dict(environ, {"DEBUG": "False"})
@patch("coltrane._configure_settings")
def test_debug_env(_configure_settings):
    initialize()

    expected = _get_settings_with_debug_false()
    expected["TEMPLATES"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_time_zone_default(_configure_settings):
    initialize()

    expected = _get_settings()
    expected["TIME_ZONE"] = "UTC"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_time_zone_setting(_configure_settings):
    initialize(TIME_ZONE="America/New_York")

    expected = _get_settings()
    expected["TIME_ZONE"] = "America/New_York"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch.dict(environ, {"TIME_ZONE": "America/Chicago"})
@patch("coltrane._configure_settings")
def test_time_zone_env(_configure_settings):
    initialize()

    expected = _get_settings()
    expected["TIME_ZONE"] = "America/Chicago"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs(_configure_settings):
    initialize()

    expected = _get_settings()

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs_content(_configure_settings, tmp_path: Path):
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test.txt").write_text("test")

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = tmp_path
    expected["TEMPLATES"] = ANY
    expected["STATICFILES_DIRS"] = expected["STATICFILES_DIRS"] = [tmp_path]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs_data(_configure_settings, tmp_path: Path):
    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text("{}")

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = tmp_path
    expected["TEMPLATES"] = ANY
    expected["STATICFILES_DIRS"] = [tmp_path]

    _configure_settings.assert_called_once_with(expected)


"""
{
    'BASE_DIR': PosixPath('/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0'),
    'ROOT_URLCONF': 'coltrane.urls',
    'DEBUG': True,
    'SECRET_KEY': <ANY>,
    'INSTALLED_APPS': ['django.contrib.humanize', 'django.contrib.sitemaps', 'django.contrib.staticfiles', 'django_fastdev', 'coltrane', 'django_browser_reload'],
    'CACHES': {
        'default': {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache'
        }
    },
    'MIDDLEWARE': ['coltrane.middleware.IsSecureMiddleware', 'django.middleware.security.SecurityMiddleware', 'django.middleware.gzip.GZipMiddleware', 'django.middleware.http.ConditionalGetMiddleware', 'django.middleware.common.CommonMiddleware', 'django.middleware.csrf.CsrfViewMiddleware', 'django.middleware.clickjacking.XFrameOptionsMiddleware', 'django_browser_reload.middleware.BrowserReloadMiddleware'],
    'TEMPLATES': <ANY>,
    'INTERNAL_IPS': [],
    'ALLOWED_HOSTS': [],
    'STATIC_ROOT': <ANY>,
    'STATIC_URL': <ANY>,
    'STATICFILES_DIRS': [PosixPath('/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0/static'), PosixPath('/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0/data')],
    'TIME_ZONE': <ANY>,
    'LOGGING': <ANY>,
    'COLTRANE': {
        'TITLE': '',
        'DESCRIPTION': '',
        'SITE_URL': '',
        'MARKDOWN_RENDERER': 'mistune',
        'MISTUNE_PLUGINS': ['strikethrough', 'footnotes', 'table', 'task_lists', 'def_list', 'abbr', 'mark', 'insert', 'superscript', 'subscript'],
        'EXTRA_FILE_NAMES': [],
        'CONTENT_DIRECTORY': 'content',
        'DATA_DIRECTORY': 'data',
        'DISABLE_WILDCARD_TEMPLATES': False,
        'IS_SECURE': False,
        'DATA_JSON5': False,
        'SITES': {}
    },
    'SETTINGS_MODULE': 'coltrane',
    'ENV': <ANY>
}
"""

"""
{
  "BASE_DIR": "/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0",
  "ROOT_URLCONF": "coltrane.urls",
  "DEBUG": true,
  "SECRET_KEY": null,
  "INSTALLED_APPS": [
    "django.contrib.humanize",
    "django.contrib.sitemaps",
    "django.contrib.staticfiles",
    "django_fastdev",
    "coltrane",
    "django_browser_reload"
  ],
  "CACHES": {
    "default": {
      "BACKEND": "django.core.cache.backends.dummy.DummyCache"
    }
  },
  "MIDDLEWARE": [
    "coltrane.middleware.IsSecureMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.gzip.GZipMiddleware",
    "django.middleware.http.ConditionalGetMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_browser_reload.middleware.BrowserReloadMiddleware"
  ],
  "TEMPLATES": [
    {
      "BACKEND": "django.template.backends.django.DjangoTemplates",
      "DIRS": [
        "/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0/templates"
      ],
      "OPTIONS": {
        "builtins": [
          "django.contrib.humanize.templatetags.humanize",
          "django.templatetags.static",
          "coltrane.templatetags.coltrane_tags"
        ],
        "context_processors": [
          "django.template.context_processors.request",
          "django.template.context_processors.debug",
          "django.template.context_processors.static",
          "coltrane.context_processors.coltrane"
        ],
        "loaders": [
          [
            "django.template.loaders.cached.Loader",
            [
              "dj_angles.template_loader.Loader",
              "django.template.loaders.filesystem.Loader",
              "django.template.loaders.app_directories.Loader"
            ]
          ]
        ]
      }
    }
  ],
  "INTERNAL_IPS": [],
  "ALLOWED_HOSTS": [],
  "STATIC_ROOT": "/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0/output/static",
  "STATIC_URL": "static/",
  "STATICFILES_DIRS": [
    "/private/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/pytest-of-adam/pytest-956/test_staticfiles_dirs_data0"
  ],
  "TIME_ZONE": "UTC",
  "LOGGING": {
    "version": 1,
    "disable_existing_loggers": false,
    "handlers": {
      "console": {
        "class": "logging.StreamHandler"
      }
    },
    "root": {
      "handlers": ["console"],
      "level": "ERROR"
    }
  },
  "COLTRANE": {
    "TITLE": "",
    "DESCRIPTION": "",
    "SITE_URL": "",
    "MARKDOWN_RENDERER": "mistune",
    "MISTUNE_PLUGINS": [
      "strikethrough",
      "footnotes",
      "table",
      "task_lists",
      "def_list",
      "abbr",
      "mark",
      "insert",
      "superscript",
      "subscript"
    ],
    "EXTRA_FILE_NAMES": [],
    "CONTENT_DIRECTORY": "content",
    "DATA_DIRECTORY": "data",
    "DISABLE_WILDCARD_TEMPLATES": false,
    "IS_SECURE": false,
    "DATA_JSON5": false,
    "SITES": {}
  },
  "SETTINGS_MODULE": "coltrane",
  "ENV": {
    "ARGS": "tests/init/test_initialize.py::test_staticfiles_dirs_data",
    "ASDF_DIR": "/Users/adam/.asdf",
    "COLORTERM": "truecolor",
    "COMMAND_MODE": "unix2003",
    "GHOSTTY_BIN_DIR": "/Applications/Ghostty.app/Contents/MacOS",
    "GHOSTTY_RESOURCES_DIR": "/Applications/Ghostty.app/Contents/Resources/ghostty",
    "GHOSTTY_SHELL_INTEGRATION_NO_SUDO": "1",
    "HEX": "0123456789abcdef",
    "HEXLOWER": "0123456789abcdef",
    "HEXUPPER": "0123456789ABCDEF",
    "HOME": "/Users/adam",
    "HOMEBREW_CELLAR": "/opt/homebrew/Cellar",
    "HOMEBREW_PREFIX": "/opt/homebrew",
    "HOMEBREW_REPOSITORY": "/opt/homebrew",
    "INFOPATH": "/opt/homebrew/share/info",
    "LANG": "en_US.UTF-8",
    "LOGNAME": "adam",
    "MANPATH": ":/usr/share/man:/usr/local/share/man::/Applications/Ghostty.app/Contents/Resources/ghostty/../man",
    "NIX_PROFILES": "/nix/var/nix/profiles/default /Users/adam/.nix-profile",
    "NIX_SSL_CERT_FILE": "/nix/var/nix/profiles/default/etc/ssl/certs/ca-bundle.crt",
    "NVM_BIN": "/Users/adam/.nvm/versions/node/v7.2.1/bin/node",
    "PATH": "/Users/adam/Source/adamghill/coltrane/.venv/bin:/opt/homebrew/bin:/opt/homebrew/sbin:/Users/adam/.nvm/versions/node/v7.2.1/bin:/Users/adam/.asdf/shims:/Users/adam/.asdf/bin:/Users/adam/.pyenv/shims:/Users/adam/.cargo/bin:/Users/adam/.nix-profile/bin:/nix/var/nix/profiles/default/bin:/Users/adam/.codeium/windsurf/bin:/usr/local/bin:/System/Cryptexes/App/usr/bin:/usr/bin:/bin:/usr/sbin:/sbin:/var/run/com.apple.security.cryptexd/codex.system/bootstrap/usr/local/bin:/var/run/com.apple.security.cryptexd/codex.system/bootstrap/usr/bin:/var/run/com.apple.security.cryptexd/codex.system/bootstrap/usr/appleinternal/bin:/Library/Apple/usr/bin:/usr/local/go/bin:/Applications/Ghostty.app/Contents/MacOS:/Users/adam/.local/bin",
    "PWD": "/Users/adam/Source/adamghill/coltrane",
    "PYENV_ROOT": "/Users/adam/.pyenv",
    "PYENV_SHELL": "fish",
    "PYTHONDONTWRITEBYTECODE": "1",
    "SHELL": "/opt/homebrew/bin/fish",
    "SHLVL": "2",
    "SSH_AUTH_SOCK": "/private/tmp/com.apple.launchd.opfPALd2Y5/Listeners",
    "STARSHIP_SESSION_KEY": "2439117425178122",
    "STARSHIP_SHELL": "fish",
    "TERM": "xterm-ghostty",
    "TERMINFO": "/Applications/Ghostty.app/Contents/Resources/terminfo",
    "TERM_PROGRAM": "ghostty",
    "TERM_PROGRAM_VERSION": "1.0.0",
    "TMPDIR": "/var/folders/4_/41j53rw1683_y0qzt18qrn4r0000gn/T/",
    "USER": "adam",
    "VIRTUAL_ENV": "/Users/adam/Source/adamghill/coltrane/.venv",
    "XDG_DATA_DIRS": "/usr/local/share:/usr/share:/Applications/Ghostty.app/Contents/Resources/ghostty/..",
    "XPC_FLAGS": "0x0",
    "XPC_SERVICE_NAME": "0",
    "_": "/Users/adam/.cargo/bin/uv",
    "__CFBundleIdentifier": "com.mitchellh.ghostty",
    "__CF_USER_TEXT_ENCODING": "0x1F5:0x0:0x0",
    "src": "coltrane",
    "PYTEST_VERSION": "8.3.3",
    "PYTEST_CURRENT_TEST": "tests/init/test_initialize.py::test_staticfiles_dirs_data (call)"
  }
}
"""



from typing import Dict, List, Optional, Union

from django import template
from django.core.handlers.wsgi import WSGIRequest
from django.http import Http404
from django.template.base import Node, Template, TextNode, token_kwargs
from django.template.exceptions import TemplateSyntaxError
from django.template.loader_tags import BLOCK_CONTEXT_KEY, BlockContext, BlockNode, construct_relative_path
from django.templatetags.static import StaticNode
from django.utils.safestring import SafeString, mark_safe

from coltrane.config.settings import get_config
from coltrane.renderer import MarkdownRenderer
from coltrane.retriever import get_content_directory, get_content_paths

register = template.Library()


class NoParentError(Exception):
    pass


def _is_content_slug_in_string(content_slug: str, slugs: Optional[str]) -> bool:
    """
    Whether a content slug is included in a string. Handles if `string` is
    comma-delimited list of slugs. Also handles any individual slug
    to check having a forward-slash prefix.
    """

    if not slugs:
        return False

    if not isinstance(slugs, str):
        raise TypeError("Slugs must be a string")

    split_slugs = slugs.split(",")

    for slug in split_slugs:
        slug_to_check = slug.strip()

        if slug_to_check.startswith("/"):
            slug_to_check = slug_to_check[1:]

        if slug_to_check == content_slug:
            return True

    return False


@register.simple_tag(takes_context=True)
def directory_contents(
    context, directory: Optional[str] = None, exclude: Optional[str] = None, order_by=None
) -> List[Dict[str, str]]:
    """
    Returns a list of content metadata for a particular directory. Useful for
    listing links to content.
    """

    request = context["request"]

    site = get_config().get_site(request)

    if not directory:
        directory = request.path
    elif isinstance(directory, SafeString):
        # Force SafeString to be a normal string so it can be used with `Path` later
        directory = directory + ""

    if directory and directory.startswith("/"):
        directory = directory[1:]

    content_paths = get_content_paths(request, str(directory))
    contents = []

    for path in content_paths:
        if path.name != "index.md":
            path_slug = path.name[:-3]
            content_slug = path_slug
            content_directory = get_content_directory()

            if directory:
                content_slug = f"{directory}/{path_slug}"
                content_directory = content_directory / directory

            if _is_content_slug_in_string(content_slug=content_slug, slugs=exclude):
                continue

            (_, metadata) = MarkdownRenderer.instance().get_html_and_markdown(content_slug, site)

            contents.append(metadata)

    if order_by and contents:
        is_reverse = False

        if order_by[0] == "-":
            is_reverse = order_by[0] == "-"
            order_by = order_by[1:]

        def _directory_content_sorter(_metadata: Dict) -> str:
            value = _metadata.get(order_by, "") or ""

            return str(value)

        contents.sort(key=_directory_content_sorter, reverse=is_reverse)

    return contents


@register.filter()
def parent(path: Union[str, WSGIRequest] = "") -> str:
    """
    Gets the the directory above `path`.
    """

    if isinstance(path, WSGIRequest) or hasattr(path, "path"):
        # Handle if a `request` is passed in
        path = path.path

    path = path.strip()

    if path.endswith("/"):
        path = path[:-1]

    if path == "":
        raise NoParentError()

    last_slash_index = path.rindex("/")
    path = path[:last_slash_index]

    return path


class IncludeMarkdownNode(Node):
    """
    Based on: `django.template.loader_tags.IncludeNode`.
    """

    context_key = "__include_markdown_context"

    def __init__(self, template, *args, **kwargs):
        self.template = template
        super().__init__(*args, **kwargs)

    def render(self, context):
        """
        Render the specified template and context.
        """

        template_name = self.template.resolve(context)

        # If the current request is for a custom site, target that site's template folder
        if request := context.get("request"):
            template_name = get_config().get_site(request).get_template_name(template_name, verify=False)

        cache = context.render_context.dicts[0].setdefault(self, {})
        template = cache.get(template_name)

        if template is None:
            template = context.template.engine.select_template((template_name,))
            cache[template_name] = template

        (html, metadata) = MarkdownRenderer.instance().render_markdown_path(template.origin.name)

        for c in context:
            for key, value in c.items():
                if key not in metadata:
                    metadata[key] = value

        return MarkdownRenderer.instance().render_html_with_django(html, metadata)


@register.tag("include_md")
def do_include_md(parser, token):
    """
    Load a markdown template and render it with the current context.

    Based on: `django.template.loader_tags.do_include`.

    Example:
        {% include_md "foo/some_include" %}
    """

    bits = token.split_contents()

    if len(bits) < 2:  # noqa: PLR2004
        raise TemplateSyntaxError(
            "%r tag takes at least one argument: the name of the template to be included." % bits[0]
        )

    bits[1] = construct_relative_path(parser.origin.template_name, bits[1])

    return IncludeMarkdownNode(
        parser.compile_filter(bits[1]),
    )


class IncludeNode(Node):
    context_key = "__include_context"

    def __init__(self, template, *args, extra_context=None, isolated_context=False, **kwargs):
        self.template = template
        self.extra_context = extra_context or {}
        self.isolated_context = isolated_context
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__qualname__}: template={self.template!r}>"

    def render(self, context):
        """
        Render the specified template and context. Cache the template object
        in render_context to avoid reparsing and loading when used in a for
        loop.
        """

        template = self.template.resolve(context)

        if isinstance(template, str) or isinstance(template, SafeString):
            if request := context.get("request"):
                template = get_config().get_site(request).get_template_name(template)

        # Does this quack like a Template?
        if not callable(getattr(template, "render", None)):
            # If not, try the cache and select_template().
            template_name = template or ()

            if isinstance(template_name, str):
                template_name = (
                    construct_relative_path(
                        self.origin.template_name,
                        template_name,
                    ),
                )
            else:
                template_name = tuple(template_name)

            cache = context.render_context.dicts[0].setdefault(self, {})
            template = cache.get(template_name)

            if template is None:
                template = context.template.engine.select_template(template_name)
                cache[template_name] = template

        # Use the base.Template of a backends.django.Template.
        elif hasattr(template, "template"):
            template = template.template

        values = {name: var.resolve(context) for name, var in self.extra_context.items()}

        if self.isolated_context:
            return template.render(context.new(values))

        with context.push(**values):
            return template.render(context)


@register.tag("include")
def do_include(parser, token):
    """
    Load a template and render it with the current context. You can pass
    additional context using keyword arguments.

    Example:
        {% include "foo/some_include" %}
        {% include "foo/some_include" with bar="BAZZ!" baz="BING!" %}

    Use the `only` argument to exclude the current context when rendering
    the included template::

        {% include "foo/some_include" only %}
        {% include "foo/some_include" with bar="1" only %}
    """

    bits = token.split_contents()

    if len(bits) < 2:
        raise TemplateSyntaxError(
            "%r tag takes at least one argument: the name of the template to be included." % bits[0]
        )

    options = {}
    remaining_bits = bits[2:]

    while remaining_bits:
        option = remaining_bits.pop(0)

        if option in options:
            raise TemplateSyntaxError("The %r option was specified more than once." % option)
        if option == "with":
            value = token_kwargs(remaining_bits, parser, support_legacy=False)

            if not value:
                raise TemplateSyntaxError('"with" in %r tag needs at least one keyword argument.' % bits[0])
        elif option == "only":
            value = True
        else:
            raise TemplateSyntaxError("Unknown argument for %r tag: %r." % (bits[0], option))

        options[option] = value

    isolated_context = options.get("only", False)
    namemap = options.get("with", {})
    bits[1] = construct_relative_path(parser.origin.template_name, bits[1])

    return IncludeNode(
        parser.compile_filter(bits[1]),
        extra_context=namemap,
        isolated_context=isolated_context,
    )


@register.filter(takes_context=True)
def to_html(context: dict, text: str) -> str:
    """
    Converts markdown to HTML.
    """

    (html, metadata) = MarkdownRenderer.instance().render_markdown_text(text)
    rendered_html = MarkdownRenderer.instance().render_html_with_django(html, metadata, request=context["request"])

    return mark_safe(rendered_html)  # noqa: S308


@register.simple_tag
def raise_404(message: Optional[str] = None):
    """Raise a 404 with an optional message."""

    if message:
        raise Http404(message)

    raise Http404()


@register.simple_tag(takes_context=True)
def last_path(context: dict) -> str:
    """Return the last part of the `HTTPRequest` path.

    For example, if `request.path` is "/something/cool", "cool" would be returned.
    """

    request = context["request"]
    path = request.path_info.strip()

    if path.endswith("/"):
        path = path[:-1]

    return path.split("/")[-1:][0]


@register.simple_tag(takes_context=True)
def paths(context: dict) -> List[str]:
    """Return all parts of the `HTTPRequest` path.

    For example, if `request.path` is "/something/cool", ["something", "cool"] would be returned.
    """

    request = context["request"]
    path = request.path_info.strip()

    if path.startswith("/"):
        path = path[1:]

    if path.endswith("/"):
        path = path[:-1]

    _paths = path.split("/")

    if _paths == [""]:
        _paths = []

    return _paths


class ColtraneStaticNode(StaticNode):
    """Used for the custom static templatetag which knows how to deal with per-site static directory. Using
    the typical Django static template tag would require having nested directories in every static directory
    to prevent different sites from using the incorrect file.

    Allows this directory structure:
    - /sites/site1/static/styles.css
    - /sites/site2/static/styles.css

    Instead of requiring namespacing static files inside each static directory:
    - /sites/site1/static/site1/styles.css
    - /sites/site2/static/site2/styles.css
    """

    def url(self, context):
        path = self.path.resolve(context)

        if "request" in context:
            request = context["request"]

            coltrane = get_config()
            site = coltrane.get_site(request)

            path = f"static/{path}"

            if site.is_custom:
                path = f"{site.folder}/{path}"

        return self.handle_simple(path)


@register.tag("static")
def do_static(parser, token):
    """
    Coltrane's override for the regular Django static templatetag.

    Join the given path with the STATIC_URL setting.

    Usage::

        {% static path [as varname] %}

    Examples::

        {% static "myapp/css/base.css" %}
        {% static variable_with_path %}
        {% static "myapp/css/base.css" as admin_base_css %}
        {% static variable_with_path as varname %}
    """

    return ColtraneStaticNode.handle_token(parser, token)


class ExtendsNode(Node):
    must_be_first = True
    context_key = "extends_context"

    def __init__(self, nodelist, parent_name, template_dirs=None):
        self.nodelist = nodelist
        self.parent_name = parent_name
        self.template_dirs = template_dirs
        self.blocks = {n.name: n for n in nodelist.get_nodes_by_type(BlockNode)}

    def __repr__(self):
        return "<%s: extends %s>" % (self.__class__.__name__, self.parent_name.token)

    def find_template(self, template_name, context):
        """
        This is a wrapper around engine.find_template(). A history is kept in
        the render_context attribute between successive extends calls and
        passed as the skip argument. This enables extends to work recursively
        without extending the same template twice.
        """

        history = context.render_context.setdefault(
            self.context_key,
            [self.origin],
        )
        template, origin = context.template.engine.find_template(
            template_name,
            skip=history,
        )
        history.append(origin)

        return template

    def get_parent(self, context):
        parent = self.parent_name.resolve(context)

        if not parent:
            error_msg = "Invalid template name in 'extends' tag: %r." % parent

            if self.parent_name.filters or isinstance(self.parent_name.var, Variable):
                error_msg += " Got this from the '%s' variable." % self.parent_name.token

            raise TemplateSyntaxError(error_msg)

        if isinstance(parent, Template):
            # parent is a django.template.Template
            return parent

        if isinstance(getattr(parent, "template", None), Template):
            # parent is a django.template.backends.django.Template
            return parent.template

        # If the current request is for a custom site, target that site's template folder
        if request := context.get("request"):
            parent = get_config().get_site(request).get_template_name(parent, verify=False)

        return self.find_template(parent, context)

    def render(self, context):
        compiled_parent = self.get_parent(context)

        if BLOCK_CONTEXT_KEY not in context.render_context:
            context.render_context[BLOCK_CONTEXT_KEY] = BlockContext()

        block_context = context.render_context[BLOCK_CONTEXT_KEY]

        # Add the block nodes from this node to the block context
        block_context.add_blocks(self.blocks)

        # If this block's parent doesn't have an extends node it is the root,
        # and its block nodes also need to be added to the block context.
        for node in compiled_parent.nodelist:
            # The ExtendsNode has to be the first non-text node.
            if not isinstance(node, TextNode):
                if not isinstance(node, ExtendsNode):
                    blocks = {n.name: n for n in compiled_parent.nodelist.get_nodes_by_type(BlockNode)}
                    block_context.add_blocks(blocks)

                break

        # Call Template._render explicitly so the parser context stays
        # the same.
        with context.render_context.push_state(compiled_parent, isolated_context=False):
            return compiled_parent._render(context)


@register.tag("extends")
def do_extends(parser, token):
    """
    Signal that this template extends a parent template.

    This tag may be used in two ways: ``{% extends "base" %}`` (with quotes)
    uses the literal value "base" as the name of the parent template to extend,
    or ``{% extends variable %}`` uses the value of ``variable`` as either the
    name of the parent template to extend (if it evaluates to a string) or as
    the parent template itself (if it evaluates to a Template object).
    """

    bits = token.split_contents()

    if len(bits) != 2:
        raise TemplateSyntaxError("'%s' takes one argument" % bits[0])

    bits[1] = construct_relative_path(parser.origin.template_name, bits[1])
    parent_name = parser.compile_filter(bits[1])
    nodelist = parser.parse()

    if nodelist.get_nodes_by_type(ExtendsNode):
        raise TemplateSyntaxError("'%s' cannot appear more than once in the same template" % bits[0])

    return ExtendsNode(nodelist, parent_name)



import logging
import sys
from contextlib import redirect_stdout
from copy import deepcopy
from io import StringIO
from os import environ, getcwd, getenv
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from django import setup as django_setup
from django.conf import settings
from django.core.handlers.wsgi import WSGIHandler
from django.core.management import execute_from_command_line
from dotenv import load_dotenv

from coltrane.config.settings import DEFAULT_COLTRANE_SETTINGS, get_config
from coltrane.module_finder import (
    is_django_compressor_installed,
    is_django_unicorn_installed,
    is_unicorn_module_available,
    is_whitenoise_installed,
)
from coltrane.utils import dict_merge

logger = logging.getLogger(__name__)

__all__ = [
    "initialize",
    "run",
]


DEFAULT_CACHES_SETTINGS = {
    "default": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    }
}

DEFAULT_MIDDLEWARE = [
    "coltrane.middleware.IsSecureMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.gzip.GZipMiddleware",
    "django.middleware.http.ConditionalGetMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


DEFAULT_INSTALLED_APPS = [
    "django.contrib.humanize",
    "django.contrib.sitemaps",
    "django.contrib.staticfiles",
    "django_fastdev",
    "coltrane",
]

COLTRANE_SETTINGS_THAT_ARE_ARRAYS = ("EXTRA_FILE_NAMES",)
COLTRANE_SETTINGS_THAT_ARE_BOOLEANS = (
    "DISABLE_WILDCARD_TEMPLATES",
    "IS_SECURE",
    "DATA_JSON5",
)


def _get_current_command():
    if len(sys.argv) > 1:
        return sys.argv[1]


def _get_base_dir(base_dir: Optional[Union[str, Path]]) -> Path:
    """
    Gets the base directory.
    """

    if base_dir is None:
        cwd = getcwd()

        base_dir = cwd / Path("sites")

        if not base_dir.exists():
            base_dir = cwd / Path("site")

            if not base_dir.exists():
                base_dir = Path(".")

    if isinstance(base_dir, str):
        base_dir = Path(base_dir)

    return base_dir


def _merge_installed_apps(django_settings: Dict[str, Any], installed_apps: List[str]) -> List[str]:
    """
    Gets the installed apps from the passed-in settings and adds `coltrane` to it.
    """

    if "INSTALLED_APPS" in django_settings:
        installed_apps.extend(list(django_settings["INSTALLED_APPS"]))

        # Remove INSTALLED_APPS from django_settings since it overrides the default settings
        # that gets merged later
        del django_settings["INSTALLED_APPS"]

    return installed_apps


def _get_caches(django_settings: Dict[str, Any]) -> Dict:
    """Gets the configured cache. Defaults to the dummy cache."""

    caches = django_settings.get("CACHES", DEFAULT_CACHES_SETTINGS)

    if "default" not in caches:
        caches = DEFAULT_CACHES_SETTINGS

    if coltrane_cache := getenv("CACHE"):
        if coltrane_cache == "dummy":
            logging.info("Use dummy cache")

            caches["default"] = {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}
        elif coltrane_cache == "memory":
            logging.info("Use local memory cache")

            caches["default"] = {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
                "LOCATION": getenv("CACHE_LOCATION", "unique-snowflake"),
            }
        elif coltrane_cache in ["filesystem", "redis", "memcache"]:
            if cache_locations := _get_from_env("CACHE_LOCATION"):
                cache_backend = None

                if coltrane_cache == "filesystem":
                    logging.info("Use filesystem cache")

                    cache_backend = "django.core.cache.backends.filebased.FileBasedCache"
                    cache_locations = cache_locations[0]
                elif coltrane_cache == "redis":
                    logging.info("Use redis cache")

                    cache_backend = "django.core.cache.backends.redis.RedisCache"
                elif coltrane_cache == "memcache":
                    logging.info("Use memcache cache")

                    cache_backend = "django.core.cache.backends.memcached.PyMemcacheCache"

                caches["default"] = {
                    "BACKEND": cache_backend,
                    "LOCATION": cache_locations,
                }
            else:
                raise Exception(f"Missing CACHE_LOCATION: '{coltrane_cache}'")
        else:
            raise Exception(f"Unknown cache backend: '{coltrane_cache}'")

    return caches


def _get_from_env(env_name: str) -> List[str]:
    """
    Retrieves environment value that could potentially be a list of strings.
    """

    env_values = []

    if value_from_env := getenv(env_name):
        env_values = value_from_env.split(",")

    return env_values


def _set_coltrane_setting(settings: Dict, initialize_settings: Dict, setting_name: str) -> Dict:
    """
    Sets a setting on the `COLTRANE` dictionary that is in the environment or passed
    in to `initialize`. Environment takes precedence.

    For example:
    - `COLTRANE_TITLE=Awesome Blog 1` in `.env` equals `COLTRANE['TITLE'] = 'Awesome Blog 1'` in settings
    - `initialize(COLTRANE_TITLE='Awesome Blog 2')` equals `COLTRANE['TITLE'] = 'Awesome Blog 2'` in settings
    - `COLTRANE_TITLE=Awesome Blog 3` in `.env` and `initialize(COLTRANE_TITLE='Awesome Blog 4')` equals
        `COLTRANE['TITLE'] = 'Awesome Blog 3'` in settings
    """

    coltrane_setting_name = f"COLTRANE_{setting_name}"
    value = getenv(coltrane_setting_name) or initialize_settings.get(coltrane_setting_name)

    if value:
        # Make sure there is a `COLTRANE` setting
        if "COLTRANE" not in settings:
            settings["COLTRANE"] = {}

        if setting_name in COLTRANE_SETTINGS_THAT_ARE_ARRAYS:
            settings["COLTRANE"][setting_name] = value.split(",")
        elif setting_name in COLTRANE_SETTINGS_THAT_ARE_BOOLEANS:
            settings["COLTRANE"][setting_name] = False

            if isinstance(value, str):
                settings["COLTRANE"][setting_name] = value.lower() == "true"
            elif isinstance(value, bool):
                settings["COLTRANE"][setting_name] = value is True
        else:
            settings["COLTRANE"][setting_name] = value

    return settings


def _get_from_env_or_settings(django_settings: Dict, key: str, default: Any) -> Any:
    """
    Get a setting from (in precedence order) the environment or the `COLTRANE` dictionary
    in settings. Normal `get_config_settings()` method does not work because it requires
    the Django settings to be configured.
    """

    val = default

    if setting_value := django_settings.get("COLTRANE", {}).get(key):
        val = setting_value

    if environment_value := getenv(f"COLTRANE_{key}"):
        val = environment_value

    return val


def _merge_settings(base_dir: Path, django_settings: Dict[str, Any]) -> Dict[str, Any]:
    """
    Merges the passed-in settings into the default `coltrane` settings.
    Passed-in settings will override the defaults.
    """

    # Assume that `argv[1] == "build"` means that the `build`
    # management command is currently being run
    is_build_management_command = _get_current_command() == "build"

    debug = django_settings.get("DEBUG", getenv("DEBUG", "True") == "True")
    time_zone = django_settings.get("TIME_ZONE", getenv("TIME_ZONE", "UTC"))

    # django_settings.get("SITES")
    # sites = django_settings.get("COLTRANE_SITES", {})

    # for key in environ.keys():
    #     if key.startswith("COLTRANE_SITE_"):

    # if not sites:
    #     sites = {"": "*"}

    # if isinstance(sites, dict):
    #     sites = Sites(sites)

    # sites = Sites(django_settings.get("COLTRANE_SITES", {"": "*"}))

    # sites = Sites(directory=Path("sites"))

    config = get_config(base_dir)

    # staticfiles_dirs = [
    #     base_dir / "static",
    # ]

    staticfiles_dirs = []

    # TODO: if sites
    # print("add base_dir to staticfiles", base_dir)
    staticfiles_dirs.append(base_dir)

    # staticfiles_dirs.append(base_dir / "adamghill" / "static")
    # print("add to staticfiles_dirs", staticfiles_dirs)

    # if not sites.has_only_default:
    # for site in coltrane.sites:
    #     static_path = base_dir / "sites" / site.folder / "static"

    #     if static_path.exists():
    #         staticfiles_dirs.append(base_dir / "sites" / site.folder / "static")

    middleware = deepcopy(DEFAULT_MIDDLEWARE)
    installed_apps = deepcopy(DEFAULT_INSTALLED_APPS)

    if is_django_unicorn_installed():
        installed_apps.append("django_unicorn")

        if is_unicorn_module_available():
            installed_apps.append("unicorn")

    if is_whitenoise_installed():
        middleware.insert(1, "whitenoise.middleware.WhiteNoiseMiddleware")
        installed_apps.insert(0, "whitenoise.runserver_nostatic")

    if is_django_compressor_installed():
        installed_apps.append("compressor")

    if debug and not is_build_management_command:
        # Add settings required for django-browser-reload when appropriate
        middleware.append("django_browser_reload.middleware.BrowserReloadMiddleware")
        installed_apps.append("django_browser_reload")

        # Add content to "staticfiles" so django-browser-reload can monitor it
        content_directory = _get_from_env_or_settings(
            django_settings, "CONTENT_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["CONTENT_DIRECTORY"]
        )
        # content_directory_absolute = base_dir / content_directory

        # if content_directory_absolute.exists():
        #     staticfiles_dirs.append(content_directory_absolute)

        # Add data to "staticfiles" so django-browser-reload can monitor it
        data_directory = _get_from_env_or_settings(
            django_settings, "DATA_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["DATA_DIRECTORY"]
        )
        # data_directory_absolute = base_dir / data_directory

        # if data_directory_absolute.exists():
        #     staticfiles_dirs.append(data_directory_absolute)

        # if not sites.has_only_default:
        for site in config.sites:
            site_content_directory = base_dir / "sites" / site.folder / content_directory

            if site_content_directory.exists():
                staticfiles_dirs.append(site_content_directory)

            site_data_directory = base_dir / "sites" / site.folder / data_directory

            if site_data_directory.exists():
                staticfiles_dirs.append(site_data_directory)

        # staticfiles_dirs.append("config.toml")

    templates = deepcopy(config.get_templates_settings())

    default_settings = {
        "BASE_DIR": base_dir,
        "ROOT_URLCONF": "coltrane.urls",
        "DEBUG": debug,
        "SECRET_KEY": getenv("SECRET_KEY"),
        "INSTALLED_APPS": _merge_installed_apps(django_settings, installed_apps),
        "CACHES": _get_caches(django_settings),
        "MIDDLEWARE": middleware,
        "TEMPLATES": templates,
        "INTERNAL_IPS": _get_from_env("INTERNAL_IPS"),
        "ALLOWED_HOSTS": _get_from_env("ALLOWED_HOSTS"),
        "STATIC_ROOT": base_dir / "output" / "static",
        "STATIC_URL": "static/",
        "STATICFILES_DIRS": staticfiles_dirs,
        "TIME_ZONE": time_zone,
        "LOGGING": {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                },
            },
            "root": {
                "handlers": ["console"],
                "level": "ERROR",
            },
        },
        "COLTRANE": DEFAULT_COLTRANE_SETTINGS,
        # Hard-coding `SETTINGS_MODULE` even though it kind of doesn't make sense, but
        # is needed for rendering a nice 500 page on local when debugging
        "SETTINGS_MODULE": "coltrane",
    }

    # Check for `COLTRANE` settings in env variables or passed into app.initialize()
    for setting_name in DEFAULT_COLTRANE_SETTINGS.keys():
        default_settings = _set_coltrane_setting(default_settings, django_settings, setting_name)

    if is_whitenoise_installed():
        default_settings["WHITENOISE_MANIFEST_STRICT"] = False
        default_settings["STATICFILES_STORAGE"] = "whitenoise.storage.CompressedManifestStaticFilesStorage"

    if is_django_compressor_installed():
        default_settings["COMPRESS_ENABLED"] = True

        default_settings["STATICFILES_FINDERS"] = (
            "django.contrib.staticfiles.finders.FileSystemFinder",
            "django.contrib.staticfiles.finders.AppDirectoriesFinder",
            "compressor.finders.CompressorFinder",
        )

        default_settings["TEMPLATES"][0]["OPTIONS"]["builtins"].append("compressor.templatetags.compress")

        if _get_current_command() == "compress":
            default_settings["COMPRESS_OFFLINE"] = True

    # Make sure BASE_DIR is a `Path` if it got passed in
    if "BASE_DIR" in django_settings and isinstance(django_settings["BASE_DIR"], str):
        django_settings["BASE_DIR"] = Path(django_settings["BASE_DIR"])

    # Override STATIC_ROOT if the output directory is manually set
    try:
        django_settings["STATIC_ROOT"] = Path(django_settings["COLTRANE"]["OUTPUT"]["DIRECTORY"]) / "static"
    except KeyError:
        pass

    django_settings = dict_merge(default_settings, django_settings, destination_overrides_source=True)
    logger.debug(f"Merged settings: {django_settings}")

    return django_settings


def _configure_settings(django_settings: Dict[str, Any]) -> None:
    """
    Configures the settings in Django.
    """

    settings.configure(**django_settings)


def _load_environment_variables(django_settings: Dict[str, Any]) -> None:
    load_dotenv(".env")

    if "ENV" not in django_settings:
        django_settings["ENV"] = {}

    django_settings["ENV"].update(dict(environ.items()))


def initialize(**django_settings) -> WSGIHandler:
    """
    Initializes the Django static site.
    """

    base_dir = _get_base_dir(django_settings.get("BASE_DIR"))
    _load_environment_variables(django_settings=django_settings)

    # coltrane_sites = django_settings.get("COLTRANE_SITES", {})

    django_settings = _merge_settings(base_dir, django_settings)
    _configure_settings(django_settings)

    django_setup()

    return WSGIHandler()


def run() -> None:
    """
    Run the Django management command based on `argv`.
    """

    if _get_current_command() == "compress":
        try:
            from compressor.exceptions import OfflineGenerationError
        except ImportError:
            logger.error("django-compressor is not installed.")
        else:
            try:
                stdout = StringIO()

                with redirect_stdout(stdout):
                    execute_from_command_line()

                compress_stdout = stdout.getvalue().replace("Compressing... done\n", "")
                print(compress_stdout)  # noqa: T201
            except OfflineGenerationError as e:
                if "No 'compress' template tags found in templates." in e.args[0]:
                    logger.error("No compress blocks found.")
                else:
                    raise
    else:
        execute_from_command_line()



import pytest

from coltrane.renderer import MistuneMarkdownRenderer


@pytest.fixture
def markdown_renderer():
    return MistuneMarkdownRenderer()


def test_render_markdown_text(markdown_renderer):
    markdown_content = """---
title: My test markdown title
---

# {{ title }}
"""
    (content, metadata) = markdown_renderer.render_markdown_text(markdown_content)
    expected = '<h1 id="title">{{ title }}</h1>'

    assert metadata.get("title") == "My test markdown title"
    assert content.strip() == expected


def test_render_markdown_text_with_code_fence(markdown_renderer):
    markdown_content = """---
title: My test markdown title
---

# {{ title }}

```
this is a lot of code
```
"""

    (content, metadata) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<h1 id="title">{{ title }}</h1>
{% verbatim %}<pre><code>this is a lot of code
</code></pre>{% endverbatim %}
"""

    assert metadata.get("title") == "My test markdown title"
    assert content == expected


def test_render_markdown_text_with_code_fence_with_language(markdown_renderer):
    markdown_content = """---
title: My test markdown title
---

# {{ title }}

```python
def blob():
    pass
```
"""

    (content, metadata) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<h1 id="title">{{ title }}</h1>
{% verbatim %}<div class="codehilite">
<pre><span></span><code><span class="k">def</span> <span class="nf">blob</span><span class="p">():</span>
    <span class="k">pass</span>
</code></pre>
</div>{% endverbatim %}
"""

    assert metadata.get("title") == "My test markdown title"
    assert content == expected


def test_render_markdown_text_with_code_fence_react(markdown_renderer):
    markdown_content = """---
title: My test markdown title
---

# {{ title }}

```javascript
function App() {
  return (
    <div style={{padding: "16px"}}></div>
  );
}
```
"""

    (content, metadata) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<h1 id="title">{{ title }}</h1>
{% verbatim %}<div class="codehilite">
<pre><span></span><code><span class="kd">function</span><span class="w"> </span><span class="nx">App</span><span class="p">()</span><span class="w"> </span><span class="p">{</span>
<span class="w">  </span><span class="k">return</span><span class="w"> </span><span class="p">(</span>
<span class="w">    </span><span class="o">&lt;</span><span class="nx">div</span><span class="w"> </span><span class="nx">style</span><span class="o">=</span><span class="p">{{</span><span class="nx">padding</span><span class="o">:</span><span class="w"> </span><span class="s2">"16px"</span><span class="p">}}</span><span class="o">&gt;&lt;</span><span class="err">/div&gt;</span>
<span class="w">  </span><span class="p">);</span>
<span class="p">}</span>
</code></pre>
</div>{% endverbatim %}
"""  # noqa: E501

    assert metadata.get("title") == "My test markdown title"
    assert content == expected


def test_render_markdown_text_with_back_ticks(markdown_renderer):
    markdown_content = """---
title: My test markdown title
---

# {{ title }}

`this is code`
"""

    (content, metadata) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<h1 id="title">{{ title }}</h1>
<p><code>this is code</code></p>
"""

    assert metadata.get("title") == "My test markdown title"
    assert content == expected


def test_render_markdown_text_with_abbr(markdown_renderer):
    markdown_content = """
The HTML specification
is maintained by the W3C.

*[HTML]: Hyper Text Markup Language
*[W3C]: World Wide Web Consortium
"""

    (content, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p>The <abbr title="Hyper Text Markup Language">HTML</abbr> specification
is maintained by the <abbr title="World Wide Web Consortium">W3C</abbr>.</p>
"""

    assert content == expected


def test_render_href_with_django_template_language_with_spaces(markdown_renderer):
    markdown_content = """
[{{ link_name }}]({{ link_href }})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><a href="{{ link_href }}">{{ link_name }}</a></p>
"""

    assert actual == expected


def test_render_href_with_django_template_language_without_spaces(markdown_renderer):
    markdown_content = """
[{{ link_name }}]({{link_href}})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><a href="{{ link_href }}">{{ link_name }}</a></p>
"""

    assert actual == expected


def test_render_img_with_django_template_language(markdown_renderer):
    markdown_content = """
![{{ image_alt }}]({% static 'images/test.jpg' %})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><img src="{% static 'images/test.jpg' %}" alt="{{ image_alt }}"/></p>
"""

    assert actual == expected


def test_render_img_with_django_template_language_with_spaces(markdown_renderer):
    markdown_content = """
![{{ image_alt }}]({% static 'images/test.jpg' 'more' %})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><img src="{% static 'images/test.jpg' 'more' %}" alt="{{ image_alt }}"/></p>
"""

    assert actual == expected


def test_render_img_with_django_template_language_with_spaces_double_quotes(
    markdown_renderer,
):
    markdown_content = """
![{{ image_alt }}]({% static "images/test.jpg" "more" %})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><img src="{% static 'images/test.jpg' 'more' %}" alt="{{ image_alt }}"/></p>
"""

    assert actual == expected


def test_render_img_with_django_template_language_with_kwarg(markdown_renderer):
    markdown_content = """
![{{ image_alt }}]({% static 'images/test.jpg' more={{ something }} %})
"""

    (actual, _) = markdown_renderer.render_markdown_text(markdown_content)
    expected = """<p><img src="{% static 'images/test.jpg' more={{ something }} %}" alt="{{ image_alt }}"/></p>
"""

    assert actual == expected



from os import environ

import pytest

from coltrane.config.coltrane import Config, Site


@pytest.fixture
def env():
    # Store the original environment
    _original_environ = environ.copy()

    # Runs the test
    try:
        yield
    except:  # noqa: E722, S110
        pass
    finally:
        # Set the environment back to the original
        environ.clear()
        environ.update(_original_environ)


@pytest.fixture
def default_site() -> Site:
    config = Config()

    return config.sites[0]



from coltrane.templatetags.coltrane_tags import to_html


def test_to_html(request):
    text = """---
title: My test markdown title
---

# {{ title }}
"""
    html = to_html(context={"request": request}, text=text)
    assert html.strip() == '<h1 id="title">My test markdown title</h1>'



from coltrane.wildcard_templates import _sort_potential_templates, get_potential_wildcard_templates


def test_get_potential_wildcard_templates():
    expected = [
        "*.html",
    ]

    slug = "test"
    actual = get_potential_wildcard_templates(slug)

    assert expected == actual


def test_get_potential_wildcard_templates_sub_directory():
    expected = ["test/*.html", "*/this.html", "*/*.html"]

    slug = "test/this"
    actual = get_potential_wildcard_templates(slug)

    assert expected == actual


def test_get_potential_wildcard_templates_sub_directories():
    expected = [
        "test/this/*.html",
        "test/*/now.html",
        "*/this/now.html",
        "test/*/*.html",
        "*/this/*.html",
        "*/*/*.html",
    ]

    slug = "test/this/now"
    actual = get_potential_wildcard_templates(slug)

    assert expected == actual


def test_sort_potential_templates():
    expected = [
        "test/this/*.html",
        "test/*/now.html",
        "test/*/*.html",
        "*/this/now.html",
        "*/this/*.html",
        "*/*/*.html",
    ]

    actual = _sort_potential_templates(
        [
            "*/*/*.html",
            "test/*/*.html",
            "*/this/now.html",
            "test/this/*.html",
            "*/this/*.html",
            "test/*/now.html",
        ]
    )

    assert expected == actual



from copy import deepcopy
from pathlib import Path
from unittest.mock import call, patch

from coltrane.config.settings import get_config


def _setup_settings(settings, tmp_path: Path) -> None:
    settings.BASE_DIR = tmp_path

    config = get_config()
    assert config.has_custom_sites is False

    # Set Django templates settings
    settings.TEMPLATES = deepcopy(config.get_templates_settings())


def test_404(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()

    response = client.get("/")
    assert response.status_code == 404


def test_404_directory_parent_traversal(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()

    response = client.get("/../")
    assert response.status_code == 404


def test_404_directory_home_traversal(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()

    response = client.get("/../")
    assert response.status_code == 404


def test_index_with_slash(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "index.md").write_text("# index")

    response = client.get("/")
    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="index">index</h1>' in actual


def test_index_without_slash(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "index.md").write_text("# index")

    response = client.get("")
    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="index">index</h1>' in actual


def test_directory_index_with_slash(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content/dir").mkdir(parents=True)
    (tmp_path / "content/dir/index.md").write_text("# dir")

    response = client.get("/dir/")
    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="dir">dir</h1>' in actual


def test_directory_index_without_slash(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content/dir").mkdir(parents=True)
    (tmp_path / "content/dir/index.md").write_text("# dir")

    response = client.get("/dir")
    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="dir">dir</h1>' in actual


def test_url_slug(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-this.md").write_text("# test this")

    response = client.get("/test-this")
    assert response.status_code == 200

    actual = response.content.decode()
    assert '\n<h1 id="test-this">test this</h1>\n\n' in actual


def test_url_slug_with_json_data(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-this-data.md").write_text("test data {{ data.test }}")
    (tmp_path / "data").mkdir()
    (tmp_path / "data/test.json").write_text("1")

    response = client.get("/test-this-data")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "\n<p>test data 1</p>\n\n" in actual


def test_url_slug_with_json5_data(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-this-data.md").write_text("test data {{ data.test }}")
    (tmp_path / "data").mkdir()
    (tmp_path / "data/test.json5").write_text("1")

    response = client.get("/test-this-data")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "\n<p>test data 1</p>\n\n" in actual


def test_url_slug_cache_headers(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE = {"VIEW_CACHE": {"SECONDS": 15}}

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-this-cache.md").write_text("test cache")

    response = client.get("/test-this-cache")
    assert response.status_code == 200
    assert response.headers.get("Cache-Control") == "max-age=15"
    assert response.headers.get("Expires")
    original_expires = response.headers.get("Expires")

    # Call view again
    response = client.get("/test-this-cache")
    assert response.status_code == 200
    assert response.headers.get("Expires") == original_expires


def test_url_slug_cache(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE = {"VIEW_CACHE": {"SECONDS": 15}}

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-this-cache.md").write_text("test cache")

    context = {
        "template": "coltrane/content.html",
        "data": {},
        "content": "<p>test cache</p>\n",
    }

    with patch(
        "coltrane.views._get_from_cache_if_enabled",
        return_value=("coltrane/content.html", context),
    ) as _get_from_cache_if_enabled:
        with patch("coltrane.views._set_in_cache_if_enabled") as _set_in_cache_if_enabled:
            response = client.get("/test-this-cache")
            assert response.status_code == 200
            assert response.headers.get("Cache-Control") == "max-age=15"
            assert response.headers.get("Expires")
            original_expires = response.headers.get("Expires")

            # Call view again
            response = client.get("/test-this-cache")
            assert response.status_code == 200
            assert response.headers.get("Expires") == original_expires

            _get_from_cache_if_enabled.assert_has_calls(calls=[call("test-this-cache"), call("test-this-cache")])
            _set_in_cache_if_enabled.assert_not_called()


def test_url_slug_direct_template(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates").mkdir(parents=True)
    (tmp_path / "templates" / "test-this.html").write_text("test this")

    response = client.get("/test-this")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "test this" in actual


def test_url_slug_direct_template_wildcard(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates").mkdir(parents=True)
    (tmp_path / "templates" / "*.html").write_text("asterisk")

    response = client.get("/test-this-missing-thing")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "asterisk" in actual


def test_url_wildcards(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates" / "test-this").mkdir(parents=True)
    (tmp_path / "templates" / "test-this" / "*.html").write_text("directory asterisk")

    response = client.get("/test-this/missing-thing")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "directory asterisk" in actual


def test_url_wildcards_404(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates").mkdir(parents=True)
    (tmp_path / "templates" / "*.html").write_text("multiple asterisk")

    response = client.get("/something/test-this-should-404")
    assert response.status_code == 404


def test_url_multiple_wildcards(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates" / "*").mkdir(parents=True)
    (tmp_path / "templates" / "*" / "*.html").write_text("multiple asterisk")

    response = client.get("/something/test-this-should-200")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "multiple asterisk" in actual


def test_url_multiple_wildcards_and_multiple_subdirectories(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (tmp_path / "templates" / "new" / "another" / "*").mkdir(parents=True)
    (tmp_path / "templates" / "new" / "another" / "*" / "*.html").write_text("multiple asterisk in subdirectory")

    response = client.get("/new/another/something/test-this-should-200")
    assert response.status_code == 200

    actual = response.content.decode()
    assert "multiple asterisk in subdirectory" in actual


def test_include(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    (settings.BASE_DIR / "templates").mkdir(parents=True)
    (settings.BASE_DIR / "templates/_partial.html").write_text("""
This is a partial
""")
    (settings.BASE_DIR / "templates/include-1.html").write_text("""
template 1

{% include '_partial.html' %}
""")

    response = client.get("/include-1")

    assert response.status_code == 200

    actual = response.content.decode()
    assert "This is a partial" in actual


def test_static(client, settings, tmp_path: Path):
    _setup_settings(settings, tmp_path)

    text = """
static 1

<img src="{% static 'img/example.png' %}" />
"""
    (settings.BASE_DIR / "templates").mkdir(parents=True)
    (settings.BASE_DIR / "templates/static-1.html").write_text(text)

    actual = client.get("/static-1")

    assert 200 == actual.status_code
    assert '<img src="static/img/example.png" />' in actual.content.decode()



from datetime import date, datetime
from pathlib import Path
from unittest.mock import patch

import pytest
from zoneinfo import ZoneInfo

from coltrane.renderer import MistuneMarkdownRenderer
from tests.fixtures import default_site


@pytest.fixture
def markdown_renderer():
    return MistuneMarkdownRenderer()


@patch(
    "coltrane.renderer.MistuneMarkdownRenderer._get_markdown_content_as_html",
    return_value=("some-content", None),
)
def test_handle_none_metadata(_get_markdown_content_as_html, markdown_renderer, default_site):
    markdown_renderer.get_html_and_markdown("some-slug", site=default_site)


def test_publish_date_datetime_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path
    settings.TIME_ZONE = "UTC"

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
publish_date: 2022-02-26 10:26:02
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["publish_date"], datetime)
    assert metadata["publish_date"] == datetime(2022, 2, 26, 10, 26, 2, 0, tzinfo=ZoneInfo(key="UTC"))


def test_publish_date_date_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path
    settings.TIME_ZONE = "UTC"

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
publish_date: 2022-02-26
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["publish_date"], date)
    assert metadata["publish_date"] == datetime(2022, 2, 26, 0, 0, 0, tzinfo=ZoneInfo(key="UTC"))


def test_draft_true_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
draft: true
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["draft"], bool)
    assert metadata["draft"]


def test_draft_false_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
draft: false
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["draft"], bool)
    assert not metadata["draft"]


def test_draft_string_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
draft: blob
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["draft"], bool)
    assert not metadata["draft"]


def test_draft_1_in_metadata(markdown_renderer, settings, tmp_path: Path, default_site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
draft: 1
---
"""
    )

    (_, metadata) = markdown_renderer.get_html_and_markdown("test-1", site=default_site)

    assert isinstance(metadata["draft"], bool)
    assert not metadata["draft"]



from coltrane.config.paths import get_output_static_directory


def test_static_root(settings):
    settings.STATIC_ROOT = "test-root"
    actual = get_output_static_directory()

    assert str(actual) == "test-root"



import codecs
import logging
import re
from dataclasses import dataclass, field
from html import unescape
from pathlib import Path
from typing import Dict, Optional, Tuple, Union
from urllib.parse import unquote

from django.conf import settings
from django.http import HttpRequest
from django.template import engines
from django.utils.html import mark_safe  # type: ignore
from django.utils.text import slugify
from django.utils.timezone import now

from coltrane.config.coltrane import Site
from coltrane.config.paths import get_content_directory
from coltrane.config.settings import (
    get_config,
    get_markdown_renderer,
    get_mistune_plugins,
    get_site_url,
)
from coltrane.retriever import get_data
from coltrane.utils import convert_to_datetime

logger = logging.getLogger(__name__)

DEFAULT_TEMPLATE = "coltrane/content.html"

SPACE_REPLACEMENT = "DJANGO-TEMPLATE-TAG-SPACE"


@dataclass
class StaticRequest(HttpRequest):
    """
    Used to mock an HttpRequest when generating the HTML for static sites.

    Required for `coltrane.templatetags.coltrane.current_direct
    """

    path: str
    META: Dict = field(default_factory=dict)
    GET: Dict = field(default_factory=dict)

    def __init__(self, path: str, meta=None, get=None):
        self.path = path
        self.META = meta or {}
        self.GET = get or {}

    @property
    def site_url(self):
        site_url = get_site_url()

        if not site_url:
            raise AssertionError("COLTRANE_SITE_URL in .env or COLTRANE.SITE_URL in settings file is required")

        return site_url

    @property
    def scheme(self) -> str:
        return str(self.site_url.split("://")[0])

    def get_host(self) -> str:
        return str(self.site_url.split("://")[1])

    def is_secure(self) -> bool:
        return self.path.startswith("https://")


class MarkdownRenderer:
    _instance = None

    def _get_markdown_content_as_html(self, slug: str, site: Site) -> Tuple[str, Optional[Dict]]:
        """
        Converts markdown file based on the slug into HTML.
        """

        path = get_content_directory(site) / f"{slug}.md"

        # print("test---path", path)

        return self.render_markdown_path(path)

    def pre_process_markdown(self, text: str) -> str:
        # Wrap code fences with Django `verbatim` templatetag; these get removed in
        # `post_process_html`
        text = re.sub(
            pattern=r"```.*?```",
            repl="{% verbatim %}\n\\g<0>\n{% endverbatim %}",
            string=text,
            flags=re.RegexFlag.DOTALL,
        )

        return text

    def post_process_html(self, html: str) -> str:
        # Remove `p` tags that get added to the `verbatim` templatetag
        html = html.replace("<p>{% verbatim %}</p>\n", "{% verbatim %}")
        html = html.replace("\n<p>{% endverbatim %}</p>", "{% endverbatim %}")

        return html

    def render_markdown_path(self, path) -> Tuple[str, Dict]:
        """
        Renders the markdown file located at path.
        """

        with codecs.open(path, "r", encoding="utf-8") as f:
            text = f.read()

            return self.render_markdown_text(text)

    def render_markdown_text(self, text: str) -> Tuple[str, Dict]:  # noqa: ARG002
        raise Exception("Missing render_markdown_text")

    def render_html_with_django(self, html: str, context: Dict, request: HttpRequest = None) -> str:
        """
        Takes the rendered HTML from the markdown and use Django to fill in any template
        variables from the `context` dictionary.
        """

        django_engine = engines["django"]
        template = django_engine.from_string(html)

        return str(template.render(context=context, request=request))

    def get_html_and_markdown(self, slug: str, site: Site) -> Tuple[str, Dict]:
        (html, metadata) = self._get_markdown_content_as_html(slug, site)

        if metadata is None:
            metadata = {}

        if "template" not in metadata:
            metadata["template"] = DEFAULT_TEMPLATE

        metadata["slug"] = slug

        return (html, metadata)

    def render_markdown(
        self,
        slug: str,
        request: Union[HttpRequest, StaticRequest],
    ) -> Tuple[str, Dict]:
        """
        Renders the markdown from the `slug` by:
        1. Rendering the markdown file into HTML
        2. Passing the HTML through Django to fill in template variables based on
            data in JSON files and markdown frontmatter

        Returns:
            Tuple of template file name (i.e. `coltrane/content.html`) and context
            dictionary.
        """

        site = get_config(base_dir=Path(".")).get_site(request)
        # print("testtt-site", site)
        # print("GOT SITE WHEN RENDER", site)
        assert site, "Missing site"

        (html, metadata) = self.get_html_and_markdown(slug, site)
        context = {"debug": settings.DEBUG, "coltrane": settings.COLTRANE}

        # Start with any metadata from the markdown frontmatter
        context.update(metadata)

        # Add JSON data to the context
        data = get_data(site=site)
        context["data"] = data

        if request:
            context["request"] = request

        # Add rendered content to the context
        content = self.render_html_with_django(html, context, request)

        context["content"] = mark_safe(content)  # noqa: S308
        template = context["template"]

        return (template, context)

    @classmethod
    def instance(cls) -> "MistuneMarkdownRenderer":
        if cls._instance is None:
            markdown_renderer = get_markdown_renderer()

            if markdown_renderer == "mistune":
                cls._instance = MistuneMarkdownRenderer()
            else:
                raise AssertionError("Invalid markdown renderer")

        return cls._instance


class MistuneMarkdownRenderer(MarkdownRenderer):
    """
    A markdown renderer for `mistune`.
    """

    from mistune.renderers.html import HTMLRenderer

    class CustomHTMLRenderer(HTMLRenderer):
        def _color_with_pygments(self, codeblock, lexer, **formatter_opts):
            import pygments
            from pygments.formatters import HtmlFormatter

            class HtmlCodeFormatter(HtmlFormatter):
                def _wrap_code(self, inner):
                    """
                    A function for use in a Pygments Formatter which wraps in <code> tags.
                    """

                    yield 0, "<code>"
                    yield from inner
                    yield 0, "</code>"

                def _add_newline(self, inner):
                    # Add newlines around the inner contents so that _strict_tag_block_re matches the outer div.
                    yield 0, "\n"
                    yield from inner
                    yield 0, "\n"

                def wrap(self, source):
                    """
                    Return the source with a code, pre, and div.
                    """

                    return self._add_newline(self._wrap_pre(self._wrap_code(source)))

            formatter_opts.setdefault("cssclass", "codehilite")
            formatter = HtmlCodeFormatter(**formatter_opts)

            return pygments.highlight(codeblock, lexer, formatter)

        def block_code(self, code: str, info=None) -> str:
            import mistune
            from pygments.lexers import get_lexer_by_name
            from pygments.util import ClassNotFound

            language = ""

            if info is not None:
                info = mistune.renderers.html.safe_entity(info.strip())
                language = info.split(None, 1)[0]

                if language:
                    try:
                        lexer = get_lexer_by_name(language)

                        return self._color_with_pygments(code, lexer)
                    except ClassNotFound:
                        pass

            return f"<pre><code>{code}</code></pre>\n"

    def __init__(self):
        import mistune
        from mistune.directives import Admonition, FencedDirective

        plugins = [*get_mistune_plugins(), FencedDirective([Admonition()])]

        self.mistune_markdown = mistune.create_markdown(
            renderer=MistuneMarkdownRenderer.CustomHTMLRenderer(),  # type: ignore
            plugins=plugins,
        )

    def _parse_and_update_metadata(self, post) -> Dict:
        """
        Add new, parse and/or cast existing values to metadata.

        `metadata["toc"]` gets generated in `_generate_toc`.
        """

        metadata = post.metadata

        if "draft" in metadata:
            if metadata["draft"] is True:
                pass
            elif metadata["draft"] == "1":
                metadata["draft"] = True
            else:
                metadata["draft"] = False

        metadata["now"] = now()

        if "publish_date" in metadata:
            metadata["publish_date"] = convert_to_datetime(metadata["publish_date"])

        return metadata

    def _generate_toc(self, content, metadata):
        """
        Update the content to add links to each header and add a `toc` key to
        `metadata` with HTML for a table of contents.
        """

        from django.utils.html import strip_tags
        from minestrone import HTML

        html = HTML(content)

        last_header_int = 0
        nested_count = 0
        strings = []

        for el in html.query("*"):
            if el.name not in ("h1", "h2", "h3", "h4", "h5", "h6"):
                continue

            header_text_slug = slugify(strip_tags(el.text))
            el.id = header_text_slug

            header_int = int(el.name[1:])

            if nested_count == 0:
                # Add the initial header
                strings.append("<ul>")
                nested_count += 1
            elif last_header_int < header_int:
                # The header is a higher number than the last one
                strings.append("<ul>")
                nested_count += 1
            elif last_header_int == header_int:
                # The header is at the same level
                strings.append("</li>")
            elif last_header_int > header_int:
                # The header is at a lower level
                for idx in range(nested_count):
                    if nested_count == 1:
                        # Last level so finish with a `li`
                        strings.append("</li>")
                    elif idx == (last_header_int - header_int):
                        # The current loop is equal to the difference
                        # between the last header and the current header
                        strings.append("</li>")
                    else:
                        # Skip down a level
                        strings.append("</li></ul>")

                    nested_count -= 1

                    if idx == (last_header_int - header_int):
                        # Break out of the loop because this iteration has gotten to the
                        # correct level
                        break

                nested_count += 1

            strings.append(f'<li><a href="#{header_text_slug}">{el.text}</a>')
            last_header_int = header_int

        # Close the correct number of `li` and `ul` tags at the end
        for _ in range(nested_count):
            strings.append("</li></ul>")

        metadata["toc"] = None

        if strings:
            toc_html = "".join(strings)
            metadata["toc"] = mark_safe(toc_html)

        content = str(html)

        return (content, metadata)

    def pre_process_markdown(self, text: str) -> str:
        text = super().pre_process_markdown(text)

        # Replace Django template variables in markdown with a marker for later because
        # mistune won't match markdown commands if there are empty strings
        text = re.sub(
            pattern=r"\({{\s*(\S+)\s*}}\)",
            repl=r"(DJANGO-TEMPLATE-VARIABLE-BEGIN-\g<1>-DJANGO-TEMPLATE-VARIABLE-END)",
            string=text,
            flags=re.RegexFlag.DOTALL,
        )

        # Get all matches for Django template tags for later adjustment
        pattern = r"\({%\s*(\S+)\s*(.*?)\s*%}\)"
        template_tag_matches = re.findall(pattern=pattern, string=text, flags=re.RegexFlag.DOTALL)

        # Replace Django template tags in markdown with a marker for later because
        # mistune won't match markdown commands if there are empty strings
        text = re.sub(
            pattern=pattern,
            repl=rf"(DJANGO-TEMPLATE-TAG-BEGIN-\g<1>{SPACE_REPLACEMENT}\g<2>-DJANGO-TEMPLATE-TAG-END)",
            string=text,
            flags=re.RegexFlag.DOTALL,
        )

        # Replace spaces between the Django template tag args/kwargs with something that
        # can be removed later
        for match in template_tag_matches:
            # Replace empty string with `SPACE_REPLACEMENT`
            # Replace double quote with a single quote
            template_tag_args = match[1].replace(" ", SPACE_REPLACEMENT).replace('"', "'")

            text = text.replace(match[1], template_tag_args)

        return text

    def post_process_html(self, html: str) -> str:
        html = super().post_process_html(html)

        # Replace the Django template language variable marker so that Django
        # variables will work for the next stage
        html = html.replace("DJANGO-TEMPLATE-VARIABLE-BEGIN-", "{{ ").replace("-DJANGO-TEMPLATE-VARIABLE-END", " }}")

        # Replace the Django template tag variable marker so that Django
        # template tags will work for the next stage
        html = html.replace("DJANGO-TEMPLATE-TAG-BEGIN-", "{% ").replace("-DJANGO-TEMPLATE-TAG-END", " %}")

        # Fix replacement string characters done in `pre_process_markdown`
        html = html.replace(SPACE_REPLACEMENT, " ")

        # Convert strings that were urlencoded by `mistune` back to their original form
        html = unquote(html)

        return html

    def render_markdown_text(self, text: str) -> Tuple[str, Dict]:
        import frontmatter

        frontmatter_post = frontmatter.loads(text)
        content = frontmatter_post.content

        content = self.pre_process_markdown(frontmatter_post.content)
        content = self.mistune_markdown(content)
        content = unescape(content)  # type: ignore
        content = self.post_process_html(content)

        metadata = self._parse_and_update_metadata(frontmatter_post)

        (content, metadata) = self._generate_toc(content, metadata)

        return (content, metadata)



import pytest

from coltrane.templatetags.coltrane_tags import NoParentError, parent


class WSGIRequest:
    @property
    def path(self):
        return "/test2/test2"


def test_parent():
    expected = "/test1"
    actual = parent("/test1/test2")

    assert actual == expected


def test_parent_wsgi_request():
    expected = "/test2"
    actual = parent(WSGIRequest())

    assert actual == expected


def test_parent_ends_with_slash():
    expected = "/test1"
    actual = parent("/test1/test2/")

    assert actual == expected


def test_parent_get_root():
    expected = ""
    actual = parent("/test1")

    assert actual == expected


def test_parent_no_parent_exception():
    with pytest.raises(NoParentError):
        parent("/")


def test_parent_no_parent_exception2():
    with pytest.raises(NoParentError):
        parent("")



from coltrane.templatetags.coltrane_tags import last_path


class WSGIRequest:
    def __init__(self, path: str):
        self.path_info = path


def test_last_path_root():
    expected = ""

    context = {"request": WSGIRequest("/")}
    actual = last_path(context=context)

    assert expected == actual


def test_last_path_empty():
    expected = ""

    context = {"request": WSGIRequest("")}
    actual = last_path(context=context)

    assert expected == actual


def test_last_path_file():
    expected = "cool"

    context = {"request": WSGIRequest("/cool")}
    actual = last_path(context=context)

    assert expected == actual


def test_last_path_subdirectory():
    expected = "cool"

    context = {"request": WSGIRequest("/something/cool")}
    actual = last_path(context=context)

    assert expected == actual


def test_last_path_subdirectory_with_trailing_slash():
    expected = "something"

    context = {"request": WSGIRequest("/something/")}
    actual = last_path(context=context)

    assert expected == actual



from datetime import datetime, timedelta
from pathlib import Path

from dateparser.timezone_parser import StaticTzInfo
from zoneinfo import ZoneInfo

from coltrane.feeds import ContentFeed


def test_link():
    actual = ContentFeed().link(None)

    assert actual == "http://localhost"


def test_items(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")

    actual = ContentFeed().items()

    assert len(actual) == 2


def test_item_link(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")

    items = ContentFeed().items()
    actual = ContentFeed().item_link(items[0])

    assert actual == "http://localhost/test"


def test_item_link_site_with_slash(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["SITE_URL"] = "http://localhost/"

    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")

    items = ContentFeed().items()
    actual = ContentFeed().item_link(items[0])

    assert actual == "http://localhost/test"


def test_item_title(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
---

test data
"""
    )

    items = ContentFeed().items()
    actual = ContentFeed().item_title(items[0])

    assert actual == "test title"


def test_item_description(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
---

test data
"""
    )

    items = ContentFeed().items()
    actual = ContentFeed().item_description(items[0])

    assert actual == "a description"


def test_item_publish_date_datetime(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
publish_date: 2024-02-25 22:36:00
---

test data
"""
    )

    settings.TIME_ZONE = "UTC"

    items = ContentFeed().items()
    actual = ContentFeed().item_pubdate(items[0])  # type: ignore

    assert actual == datetime(2024, 2, 25, 22, 36, tzinfo=ZoneInfo(key="UTC"))


def test_item_publish_date_datetime_with_tz(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
publish_date: 2024-04-03 20:22:16 -0500
---

test data
"""
    )

    items = ContentFeed().items()
    actual = ContentFeed().item_pubdate(items[0])  # type: ignore

    tz = StaticTzInfo("UTC", timedelta(hours=-5))
    expected = datetime(2024, 4, 3, 20, 22, 16, tzinfo=tz)

    assert actual == expected


def test_item_publish_date_int(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
publish_date: 1722717836
---

test data
"""
    )

    settings.TIME_ZONE = "UTC"

    items = ContentFeed().items()
    actual = ContentFeed().item_pubdate(items[0])  # type: ignore

    assert actual == datetime(2024, 8, 3, 20, 43, 56, tzinfo=ZoneInfo(key="UTC"))


def test_item_publish_date_date(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
publish_date: 2024-02-25
---

test data
"""
    )

    settings.TIME_ZONE = "UTC"

    items = ContentFeed().items()
    actual = ContentFeed().item_pubdate(items[0])  # type: ignore

    assert actual == datetime(2024, 2, 25, 0, 0, 0, tzinfo=ZoneInfo(key="UTC"))


def test_item_publish_date_with_time_zone(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: test title
description: a description
publish_date: 2024-02-25 22:36:00
---

test data
"""
    )

    settings.TIME_ZONE = "America/Chicago"

    items = ContentFeed().items()
    actual = ContentFeed().item_pubdate(items[0])

    assert actual == datetime(2024, 2, 25, 22, 36, tzinfo=ZoneInfo(key="America/Chicago"))



from typing import Annotated, Dict, Generator, Optional, Union

import msgspec

from coltrane.config.paths import get_redirects_json


class Redirect(msgspec.Struct):
    """Data for a redirect"""

    to_url: Optional[str] = msgspec.field(name="url")
    permanent: bool = False
    from_url: str = ""


def get_redirects() -> Generator[Redirect, None, None]:
    redirects_json_path = get_redirects_json()

    if not redirects_json_path.exists():
        return

    with redirects_json_path as f:
        paths = msgspec.json.decode(f.read_bytes(), type=Annotated[Dict[str, Union[str, Redirect]], msgspec.Meta()])

        for from_url, to_url in paths.items():
            if from_url.startswith("/"):
                from_url = from_url[1:]  # noqa: PLW2901

            if isinstance(to_url, Redirect):
                redirect = to_url
                redirect.from_url = from_url
            else:
                redirect = Redirect(from_url=from_url, to_url=to_url, permanent=False)

            yield redirect



from coltrane.templatetags.coltrane_tags import paths


class WSGIRequest:
    def __init__(self, path: str):
        self.path_info = path


def test_paths_root():
    expected = []

    context = {"request": WSGIRequest("/")}
    actual = paths(context=context)

    assert expected == actual


def test_paths_empty():
    expected = []

    context = {"request": WSGIRequest("")}
    actual = paths(context=context)

    assert expected == actual


def test_paths_file():
    expected = ["cool"]

    context = {"request": WSGIRequest("/cool")}
    actual = paths(context=context)

    assert expected == actual


def test_paths_subdirectory():
    expected = ["something", "cool"]

    context = {"request": WSGIRequest("/something/cool")}
    actual = paths(context=context)

    assert expected == actual


def test_paths_subdirectory_with_trailing_slash():
    expected = ["something"]

    context = {"request": WSGIRequest("/something/")}
    actual = paths(context=context)

    assert expected == actual



from pathlib import Path

import msgspec
import pytest

from coltrane.config.redirects import get_redirects


def test_get_redirects_url_url(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "redirects.json").write_text('{"/test-1": "/test-2"}')

    redirects = list(get_redirects())
    assert redirects

    actual = redirects[0]

    assert actual.from_url == "test-1"
    assert actual.to_url == "/test-2"
    assert actual.permanent is False


def test_get_redirects_url_dict(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "redirects.json").write_text('{"/test-3": {"url": "/test-4"}}')

    redirects = list(get_redirects())
    assert redirects

    actual = redirects[0]

    assert actual.from_url == "test-3"
    assert actual.to_url == "/test-4"
    assert actual.permanent is False


def test_get_redirects_url_dict_with_permanent(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "redirects.json").write_text('{"/test-5": {"url": "/test-6", "permanent": true}}')

    redirects = list(get_redirects())
    assert redirects

    actual = redirects[0]

    assert actual.from_url == "test-5"
    assert actual.to_url == "/test-6"
    assert actual.permanent is True


def test_get_redirects_url_dict_missing_url(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "redirects.json").write_text('{"/test-5": {}}')

    with pytest.raises(msgspec.ValidationError) as e:
        list(get_redirects())

    assert "Object missing required field `url`" in e.exconly()



from copy import deepcopy
from pathlib import Path
from unittest.mock import ANY, Mock, patch

from coltrane import _merge_settings
from tests.init.test_initialize import (  # type: ignore
    DEFAULT_SETTINGS,
    _get_settings,
)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
def test_merge_settings_no_extras_with_args():
    expected = deepcopy(DEFAULT_SETTINGS)
    expected["INSTALLED_APPS"].append("django_browser_reload")
    expected["MIDDLEWARE"].append("django_browser_reload.middleware.BrowserReloadMiddleware")

    actual = _merge_settings(Path("."), {})

    # Remove the ENV key because it won't be set for this code path
    del expected["ENV"]

    assert actual == expected


@patch("coltrane.is_whitenoise_installed", Mock(return_value=True))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
def test_merge_settings_with_whitenoise():
    expected = _get_settings()

    expected.update(
        {
            "WHITENOISE_MANIFEST_STRICT": False,
            "STATICFILES_STORAGE": ANY,
        }
    )

    expected["INSTALLED_APPS"].insert(0, "whitenoise.runserver_nostatic")
    expected["MIDDLEWARE"].insert(1, "whitenoise.middleware.WhiteNoiseMiddleware")

    # Remove the ENV key because it won't be set for this code path
    del expected["ENV"]

    actual = _merge_settings(Path("."), {})

    assert actual == expected


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=True))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
def test_merge_settings_with_django_unicorn():
    actual = _merge_settings(Path("."), {})
    installed_apps = actual["INSTALLED_APPS"]

    assert "django_unicorn" in installed_apps
    assert "unicorn" not in installed_apps


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=True))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=True))
def test_merge_settings_with_django_unicorn_and_unicorn():
    actual = _merge_settings(Path("."), {})
    installed_apps = actual["INSTALLED_APPS"]

    assert "django_unicorn" in installed_apps
    assert "unicorn" in installed_apps


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
def test_merge_settings_no_django_unicorn():
    actual = _merge_settings(Path("."), {})
    installed_apps = actual["INSTALLED_APPS"]

    assert "django_unicorn" not in installed_apps
    assert "unicorn" not in installed_apps


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=True))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
def test_merge_settings_with_django_compressor():
    actual = _merge_settings(Path("."), {})

    assert "compressor" in actual["INSTALLED_APPS"]

    assert "COMPRESS_ENABLED" in actual
    assert actual["COMPRESS_ENABLED"]

    assert "STATICFILES_FINDERS" in actual

    assert "compressor.finders.CompressorFinder" in actual["STATICFILES_FINDERS"]
    assert "compressor.templatetags.compress" in actual["TEMPLATES"][0]["OPTIONS"]["builtins"]


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=True))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("sys.argv", ["app.py", "compress"])
def test_merge_settings_with_django_compressor_compress_offline():
    actual = _merge_settings(Path("."), {})

    assert "COMPRESS_OFFLINE" in actual
    assert actual["COMPRESS_OFFLINE"] is True


# @patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
# @patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
# @patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
# @patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
# @patch("coltrane.is_dj_angles_installed", Mock(return_value=True))
# def test_merge_settings_with_dj_angles():
#     actual = _merge_settings(Path("."), {})
#     assert actual["TEMPLATES"]
#     template = actual["TEMPLATES"][0]

#     assert "APP_DIRS" not in template

#     assert template["OPTIONS"]["loaders"]
#     assert template["OPTIONS"]["loaders"][0] == "dj_angles.template_loader.Loader"


# @patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
# @patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
# @patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
# @patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
# @patch("coltrane.is_dj_angles_installed", Mock(return_value=False))
# def test_merge_settings_no_dj_angles():
#     actual = _merge_settings(Path("."), {})
#     assert actual["TEMPLATES"]
#     template = actual["TEMPLATES"][0]

#     assert "APP_DIRS" in template
#     assert template["APP_DIRS"] is True

#     if "loaders" in template["OPTIONS"]:
#         assert "dj_angles.template_loader.Loader" not in template["OPTIONS"]["loaders"]



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import date, datetime
from functools import wraps
from typing import Dict, List, Optional, Union

import dateparser
from django.utils.timezone import get_current_timezone, is_naive, make_aware

logger = logging.getLogger(__name__)


def dict_merge(
    source: Dict,
    destination: Dict,
    destination_overrides_source=False,  # noqa: FBT002
    path: Optional[List[str]] = None,
) -> Dict:
    """
    Deep merge two dictionaries.

    Shamelessly swiped from https://stackoverflow.com/a/7205107.
    """

    if path is None:
        path = []

    for key in destination:
        if key in source:
            if isinstance(source[key], dict) and isinstance(destination[key], dict):
                dict_merge(
                    source=source[key],
                    destination=destination[key],
                    destination_overrides_source=destination_overrides_source,
                    path=[*path, str(key)],
                )
            elif source[key] == destination[key]:
                pass  # same leaf value
            elif destination_overrides_source:
                source[key] = destination[key]
            else:
                raise Exception("Conflict at %s" % ".".join([*path, str(key)]))
        else:
            source[key] = destination[key]

    return source


def convert_to_datetime(obj: Union[str, int, datetime, date]) -> datetime:
    """Convert different objects that could be a datetime into a datetime."""

    dt: datetime

    if isinstance(obj, datetime):
        dt = obj
    elif isinstance(obj, date):
        dt = datetime.combine(obj, datetime.min.time())
    elif isinstance(obj, str):
        dt = dateparser.parse(obj)  # type: ignore
    elif isinstance(obj, int):
        dt = datetime.fromtimestamp(obj)  # noqa: DTZ006
    else:
        raise TypeError(f"Unknown type for obj: {type(obj)!s}")

    if dt and is_naive(dt):
        dt = make_aware(dt, get_current_timezone())

    return dt


def threadpool(func):
    """
    A decorator to convert a regular function so that it gets run in another thread.

    ```python
    # does not block, returns Future object
    future = some_long_running_process()

    # this blocks, waiting for the result
    result = future.result()
    ```

    More details: https://stackoverflow.com/a/14331755
    """

    @wraps(func)
    def wrap(*args, **kwargs):
        return ThreadPoolExecutor().submit(func, *args, **kwargs)

    return wrap



"""
A little wrapper CLI over a few Django management commands.
"""

from importlib.metadata import version
from os import environ, getcwd
from pathlib import Path
from stat import S_IEXEC
from subprocess import run as run_process

import rich_click as click
from click_aliases import ClickAliasedGroup
from django.core.management.utils import get_random_secret_key
from rich_click.rich_command import RichCommand

FILES_PATH = Path(__file__).parent / "default-files"
APP_DIR = "site"


def _run_management_command(command_name, *args):
    current_directory = getcwd()
    file_path = Path(current_directory) / "app.py"

    if not file_path.exists():
        file_path = Path(current_directory) / APP_DIR / "app.py"

    if not file_path.exists():
        file_path = Path(current_directory) / "sites" / "app.py"

    if not file_path.exists():
        raise Exception("app.py could not be found.")

    run_process([file_path, command_name, *list(args)])  # noqa: S603, PLW1510


def _copy_file(file_name, in_app_dir=False) -> Path:
    default_file_name = file_name

    if file_name.startswith("."):
        default_file_name = file_name[1:]

    with (FILES_PATH / default_file_name).open() as f:
        file_text = f.read()

        if "__coltrane_version__" in file_text:
            coltrane_version = version("coltrane")
            file_text = file_text.replace("__coltrane_version__", coltrane_version)

        if "__app_name__" in file_text:
            current_directory = getcwd().split("/")
            app_name = current_directory[len(current_directory) - 1 :][0]

            file_text = file_text.replace("__app_name__", app_name)

        path = Path(file_name)

        if in_app_dir:
            path = APP_DIR / path

        path.write_text(file_text)

        return path


class AliasedCommands(ClickAliasedGroup, RichCommand):
    pass


@click.group(cls=AliasedCommands, help="Runs commands for coltrane.")
@click.version_option()
def cli():
    pass


@cli.command(help="Create a new coltrane site. Alias: init.", aliases=["init"])
@click.option("--force/--no-force", default=False, help="Force creating a new site")
def create(force):
    app_file = Path("app.py")

    if not app_file.exists():
        app_file = APP_DIR / Path("app.py")

    if app_file.exists() and not force:
        click.secho("Skip project creation because app.py already exists.", fg="red")
    else:
        click.secho("Start creating files...\n", fg="yellow")
        Path(APP_DIR).mkdir(exist_ok=True)
        (APP_DIR / Path("__init__.py")).touch()

        click.secho("- Create .gitignore")
        _copy_file(".gitignore")

        click.secho("- Create Dockerfile")
        _copy_file("Dockerfile")

        click.secho("- Create pyproject.toml")
        _copy_file("pyproject.toml")

        click.secho("- Create README.md")
        _copy_file("README.md")

        click.secho(f"- Create {APP_DIR}/.env")
        env_file = _copy_file(".env", in_app_dir=True)

        # Add randomly generated secret key
        with env_file.open("a") as f:
            f.write(f"SECRET_KEY={get_random_secret_key()}")

        click.secho(f"- Create {APP_DIR}/.watchmanconfig")
        _copy_file(".watchmanconfig", in_app_dir=True)

        click.secho(f"- Create {APP_DIR}/app.py")
        app_file = _copy_file("app.py", in_app_dir=True)

        click.secho(f"- Set {APP_DIR}/app.py as executable")
        app_file.chmod(app_file.stat().st_mode | S_IEXEC)

        click.secho(f"- Create {APP_DIR}/gunicorn.conf.py")
        _copy_file("gunicorn.conf.py", in_app_dir=True)

        click.secho("- Create site/content directory")
        (APP_DIR / Path("content")).mkdir(exist_ok=True)
        (APP_DIR / Path("content") / "index.md").write_text("# index.md")

        click.secho("- Create site/data directory")
        (APP_DIR / Path("data")).mkdir(exist_ok=True)

        click.secho("- Create site/static directory")
        (APP_DIR / Path("static")).mkdir(exist_ok=True)

        click.secho("- Create site/templates directory")
        (APP_DIR / Path("templates")).mkdir(exist_ok=True)

    click.secho()
    click.secho("For local development: ", nl=False, fg="green")
    click.secho("coltrane play")
    click.secho("Build static HTML: ", nl=False, fg="green")
    click.secho("coltrane record")


@cli.command(help="Start a local development server. Alias: serve.", aliases=["serve"])
@click.option("--config", default="coltrane.toml", help="Config file name")
@click.option("--port", default=8000, help="Port to serve localhost from")
def play(config, port):
    environ.setdefault("COLTRANE_CONFIG_FILE", config)
    _run_management_command("runserver", f"0:{port}")


@cli.command(help="Generates HTML output. Aliases: rec, build.", aliases=["rec", "build"])
@click.option("--force/--no-force", default=False, help="Force HTML generation")
@click.option("--threads", type=int, help="Number of threads to use when generating static files")
@click.option("--output", help="Output directory")
@click.option("--ignore/--no-ignore", default=False, help="Ignore errors")
def record(force, threads, output, ignore):
    args = []

    if force:
        args.append("--force")

    if output:
        args.append("--output")
        args.append(output)

    if threads:
        args.append("--threads")
        args.append(str(threads))

    if ignore:
        args.append("--ignore")

    _run_management_command("build", *args)



from pathlib import Path
from unittest.mock import patch

from coltrane.console import APP_DIR, _run_management_command


def test_run_management_command(tmp_path):
    (Path(tmp_path) / APP_DIR).mkdir(exist_ok=True)
    (Path(tmp_path) / APP_DIR / "app.py").touch()

    with patch("coltrane.console.getcwd") as getcwd:
        getcwd.return_value = str(tmp_path)

        with patch("coltrane.console.run_process") as run_process:
            _run_management_command("test-command", "arg1")

            run_process.assert_called_once_with([tmp_path / APP_DIR / "app.py", "test-command", "arg1"])

        getcwd.assert_called_once()



from pathlib import Path
from unittest.mock import patch

import pytest

from coltrane.config.settings import reset_config_cache
from coltrane.renderer import MistuneMarkdownRenderer, StaticRequest


@pytest.fixture
def markdown_renderer():
    return MistuneMarkdownRenderer()


def test_render_markdown(markdown_renderer, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-2.md").write_text(
        """---
template: test-template.html
---

test data
"""
    )

    expected_template = "test-template.html"
    expected_content = "<p>test data</p>\n"
    expected_data = {}

    static_request = StaticRequest(path="/")

    (actual_template, actual_context) = markdown_renderer.render_markdown("test-2", static_request)

    assert actual_template == expected_template
    assert actual_context.get("content") == expected_content
    assert actual_context.get("data") == expected_data
    assert actual_context.get("template") == expected_template


def test_render_markdown_has_debug(markdown_renderer, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-2.md").write_text(
        """---
template: test-template.html
---

test data |{{ debug }}|
"""
    )

    expected_template = "test-template.html"
    expected_content = "<p>test data |False|</p>\n"
    expected_data = {}

    static_request = StaticRequest(path="/")

    (actual_template, actual_context) = markdown_renderer.render_markdown("test-2", static_request)

    assert actual_template == expected_template
    assert actual_context.get("content") == expected_content
    assert actual_context.get("data") == expected_data
    assert actual_context.get("template") == expected_template


@patch(
    "coltrane.renderer.MistuneMarkdownRenderer.render_markdown_path",
    return_value=("test-content", {}),
)
def test_render_markdown_metadata(
    render_markdown_path,
    markdown_renderer,
):
    static_request = StaticRequest(path="/")

    (_, metadata) = markdown_renderer.render_markdown("test-2", static_request)
    assert metadata.get("content") == "test-content"


def test_render_markdown_mistune(markdown_renderer, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path

    reset_config_cache()

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-2.md").write_text(
        """---
template: test-template.html
---

test data
"""
    )

    expected_template = "test-template.html"
    expected_content = "<p>test data</p>\n"
    expected_data = {}

    static_request = StaticRequest(path="/")

    (actual_template, actual_context) = markdown_renderer.render_markdown("test-2", static_request)

    assert actual_template == expected_template
    assert actual_context.get("content") == expected_content
    assert actual_context.get("data") == expected_data
    assert actual_context.get("template") == expected_template



import logging
from pathlib import Path
from typing import TYPE_CHECKING

from django.template.library import InvalidTemplateLibrary, import_library

from coltrane.module_finder import is_dj_angles_installed

if TYPE_CHECKING:
    from coltrane.config.coltrane import Config

logger = logging.getLogger(__name__)


class TemplatesConfigurator:
    def __init__(self, config: "Config", *args, **kwargs):
        self.config = config

    def _get_template_tag_module_name(self, file: Path) -> str:
        """
        Get a dot notation module name if a particular file path is a template tag.
        """

        base_dir = self.config.base_dir

        # TODO: Cleaner way to convert a string path to a module dot notation?
        module_name = str(file)

        if str(base_dir) != ".":
            module_name = module_name.replace(str(base_dir), "")

        module_name = module_name.replace("/", ".")

        if module_name.startswith("."):
            module_name = module_name[1:]

        if module_name.endswith(".py"):
            module_name = module_name[:-3]
        else:
            raise InvalidTemplateLibrary()

        import_library(module_name)

        return module_name

    def get_settings(self) -> list[dict]:
        """
        Gets default template settings, including templates and built-in template tags.
        """

        template_dirs = []
        templatetags_dirs = [self.config.base_dir / "templatetags"]

        if self.config.has_custom_sites:
            # Add the base directory so that overriden `include` and `static` templatetags will
            # automatically allow per-site templates to be used.
            # This is in lieu of adding template directories for each site's template folders here
            # which would also "work", but would require templates to live in another sub-directory under
            # `templates` like a normal Django app which is sub-optimal, aka not what I wanted to happen.
            template_dirs.extend([self.config.base_dir, self.config.base_dir / "templates"])

            # Add each site's templatetag folders
            for site in self.config.sites:
                templatetags_dirs.append(self.config.base_dir / site.folder / "templatetags")
        else:
            template_dirs.append(self.config.base_dir / "templates")

        template_tags = []

        for templatetags_dir in templatetags_dirs:
            if templatetags_dir.exists():
                for template_tag_path in templatetags_dir.rglob("*.py"):
                    if template_tag_path.is_file():
                        try:
                            module_name = self._get_template_tag_module_name(template_tag_path)
                            template_tags.append(module_name)
                        except InvalidTemplateLibrary as e:
                            logger.debug(e)

        builtins = [
            "django.contrib.humanize.templatetags.humanize",
            "django.templatetags.static",
            "coltrane.templatetags.coltrane_tags",  # this needs to be last so it overrides the Django templatetags
        ]
        builtins.extend(template_tags)

        template_settings = {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
            "DIRS": template_dirs,
            "OPTIONS": {
                "builtins": builtins,
                "context_processors": [
                    "django.template.context_processors.request",
                    "django.template.context_processors.debug",
                    "django.template.context_processors.static",
                    "coltrane.context_processors.coltrane",
                ],
            },
        }

        if is_dj_angles_installed():
            del template_settings["APP_DIRS"]

            if self.config.is_debug:
                template_settings["OPTIONS"]["loaders"] = [
                    "dj_angles.template_loader.Loader",
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ]
            else:
                template_settings["OPTIONS"]["loaders"] = [
                    (
                        "django.template.loaders.cached.Loader",
                        [
                            "dj_angles.template_loader.Loader",
                            "django.template.loaders.filesystem.Loader",
                            "django.template.loaders.app_directories.Loader",
                        ],
                    )
                ]

        return [template_settings]



import json
from hashlib import md5
from io import StringIO
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from django.core.management import call_command


def _call_build_command(*args, **kwargs) -> str:
    stdout = StringIO()
    stderr = StringIO()

    call_command(
        "build",
        *args,
        stdout=stdout,
        stderr=stderr,
        **kwargs,
    )

    assert not stderr.getvalue()

    return stdout.getvalue()


def _reset_settings(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.STATIC_ROOT = tmp_path / "output" / "static"


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress")
@patch("coltrane.management.commands.build.Command._call_collectstatic")
def test_build_command(_call_collectstatic, _call_compress, settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("# test 1")

    assert not (tmp_path / "output.json").exists()

    _call_build_command()

    _call_collectstatic.assert_called_once()
    _call_compress.assert_called_once()

    assert (tmp_path / "output").exists()
    assert (tmp_path / "output" / "test-1" / "index.html").exists()
    assert (tmp_path / "output.json").exists()


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress")
@patch("coltrane.management.commands.build.Command._call_collectstatic")
def test_build_command_index_md(_call_collectstatic, _call_compress, settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "index.md").write_text("# index")

    assert not (tmp_path / "output.json").exists()

    _call_build_command()

    _call_collectstatic.assert_called_once()
    _call_compress.assert_called_once()

    assert (tmp_path / "output").exists()
    assert (tmp_path / "output" / "index.html").exists()
    assert (tmp_path / "output.json").exists()


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress")
@patch("coltrane.management.commands.build.Command._call_collectstatic")
def test_build_command_directory_index_md(_call_collectstatic, _call_compress, settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create content file
    (tmp_path / "content").mkdir()
    (tmp_path / "content/dir").mkdir()
    (tmp_path / "content/dir/index.md").write_text("# dir")

    assert not (tmp_path / "output.json").exists()

    _call_build_command()

    _call_collectstatic.assert_called_once()
    _call_compress.assert_called_once()

    assert (tmp_path / "output").exists()
    assert (tmp_path / "output/dir").exists()
    assert (tmp_path / "output/dir/index.html").exists()
    assert (tmp_path / "output.json").exists()


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
def test_build_command_updates_output_manifest(settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create output.json
    (tmp_path / "output.json").write_text("{}")

    # Create content directory
    (tmp_path / "content").mkdir()
    markdown_file = tmp_path / "content" / "test-1.md"
    markdown_file.write_text("# test 1")

    _call_build_command()

    assert (tmp_path / "output").exists()
    assert (tmp_path / "output" / "test-1" / "index.html").exists()

    mtime = markdown_file.stat().st_mtime
    file_hash = md5(markdown_file.read_bytes()).hexdigest()  # noqa: S324
    expected = '{"test-1.md": {"mtime": ' + str(mtime) + ', "md5": "' + file_hash + '"}}'
    actual = (tmp_path / "output.json").read_text()

    assert actual == expected


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
def test_build_command_force(settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    stdout = _call_build_command("--force")

    assert "Force update because " in stdout


@pytest.mark.slow
@patch("coltrane.management.commands.build.Command._call_compress", Mock())
@patch("coltrane.management.commands.build.Command._call_collectstatic", Mock())
def test_build_command_staticfiles_force(settings, tmp_path):
    _reset_settings(settings, tmp_path)

    # Create content directory
    (tmp_path / "content").mkdir()

    # Create staticfiles.json
    (tmp_path / "output").mkdir()
    (tmp_path / "output" / "static").mkdir()
    (tmp_path / "output" / "static" / "staticfiles.json").write_text("{}")

    # Fake staticfiles.json metadata
    (tmp_path / "output.json").write_text(
        json.dumps(
            {
                "staticfiles.json": {"mtime": -1, "md5": "not-a-hash"},
            }
        )
    )

    stdout = _call_build_command()

    assert "Force update because static file(s) updated" in stdout
    assert (tmp_path / "output.json").exists()



from coltrane.views import _normalize_slug


def test_normalize_slug_default():
    expected = "index"
    actual = _normalize_slug("")

    assert actual == expected


def test_normalize_slug_forward_slash():
    expected = "index"
    actual = _normalize_slug("/")

    assert actual == expected


def test_normalize_slug_both_slashes():
    expected = "test-slug"
    actual = _normalize_slug("/test-slug/")

    assert actual == expected


def test_normalize_slug_startswith_slash():
    expected = "test-slug"
    actual = _normalize_slug("/test-slug")

    assert actual == expected


def test_normalize_slug_endswith_slash():
    expected = "test-slug"
    actual = _normalize_slug("test-slug/")

    assert actual == expected


def test_normalize_slug_none():
    expected = "index"
    actual = _normalize_slug(None)

    assert actual == expected



import pytest

from coltrane.templatetags.coltrane_tags import _is_content_slug_in_string


def test_is_content_slug_in_string():
    with pytest.raises(TypeError) as e:
        _is_content_slug_in_string("content-slug", 1)  # type: ignore

    assert e.exconly() == "TypeError: Slugs must be a string"



from os import environ

from coltrane import _get_from_env_or_settings
from tests.fixtures import *  # noqa: F403


def test_get_from_env_or_settings_default():
    expected = "content1"

    settings = {}
    actual = _get_from_env_or_settings(settings, "CONTENT_DIRECTORY", "content1")

    assert expected == actual


def test_get_from_env_or_settings_in_settings():
    expected = "content2"

    settings = {"COLTRANE": {"CONTENT_DIRECTORY": "content2"}}
    actual = _get_from_env_or_settings(settings, "CONTENT_DIRECTORY", "content1")

    assert expected == actual


def test_get_from_env_or_settings_in_environment(env):
    expected = "content3"

    environ["COLTRANE_CONTENT_DIRECTORY"] = "content3"
    settings = {"COLTRANE": {"CONTENT_DIRECTORY": "content2"}}
    actual = _get_from_env_or_settings(settings, "CONTENT_DIRECTORY", "content1")

    assert expected == actual



import pytest

from coltrane.renderer import MistuneMarkdownRenderer

markdown_texts = [
    """
```python
def blob():
    pass
```
""",
    """
```javascript
function blob() {
    return 0
}
```
""",
    """
```
some code here
```
""",
    """
```unknown_some_here
some code here
```
""",
]


@pytest.mark.skip("Not needed since there is only one markdown renderer now")
@pytest.mark.parametrize("text", markdown_texts)
def test_compare_code(text):
    mistune_markdown_renderer = MistuneMarkdownRenderer()

    (
        mistune_content,
        _,
    ) = mistune_markdown_renderer.render_markdown_text(text)

    # Remove linebreak differences
    mistune_content = mistune_content.replace("\n", "")

    # assert other_content == mistune_content



from pathlib import Path
from unittest.mock import ANY

import pytest

from coltrane.config.coltrane import Site
from coltrane.renderer import MistuneMarkdownRenderer
from tests.fixtures import default_site  # noqa: F401


@pytest.fixture
def markdown_renderer():
    return MistuneMarkdownRenderer()


def test_get_markdown_content_as_html_with_frontmatter(markdown_renderer, settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """---
template: test-template.html
---

test data
"""
    )

    rendered_html = "<p>test data</p>\n"
    context = {"template": "test-template.html", "toc": ANY, "now": ANY}
    expected = (rendered_html, context)
    actual = markdown_renderer._get_markdown_content_as_html("test-1", site=default_site)

    assert actual == expected


def test_get_markdown_content_toc(markdown_renderer, settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text(
        """
# title

## test data
"""
    )

    expected_content = """<h1 id="title">title</h1>
<h2 id="test-data">test data</h2>
"""

    expected_toc = '<ul><li><a href="#title">title</a><ul><li><a href="#test-data">test data</a></li></ul></li></ul>'

    expected_metadata = {
        "now": ANY,
        "toc": expected_toc,
    }
    expected = (expected_content, expected_metadata)
    actual = markdown_renderer._get_markdown_content_as_html("test-1", site=default_site)

    assert actual == expected



from pathlib import Path
from unittest.mock import patch

from coltrane.config.paths import get_base_directory


def test_get_base_directory_no_base_dir_setting(settings):
    delattr(settings, "BASE_DIR")

    with patch("coltrane.config.paths.getcwd", return_value="1234"):
        expected = Path("1234")
        actual = get_base_directory()

        assert actual == expected


def test_get_base_directory_str_base_dir_setting(settings):
    settings.BASE_DIR = "5678"

    expected = Path("5678")
    actual = get_base_directory()

    assert actual == expected



from coltrane.config import settings


class IsSecureMiddleware:
    """Overrides request.is_secure() to return `True` based on a setting.

    Only use if you're **always** serving with HTTPS.

    Based on https://noumenal.es/notes/til/django/csrf-trusted-origins/ and
    https://gist.github.com/carltongibson/648099cd34b2c0a18e948c917a5c48fd.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not request.is_secure() and settings.get_is_secure():
            # Override `request.is_secure`
            request.is_secure = lambda: True

        response = self.get_response(request)
        return response



import pytest

from coltrane.renderer import StaticRequest


def test_static_request_scheme(settings):
    settings.COLTRANE = {"SITE_URL": "http://localhost"}
    request = StaticRequest(path="/")

    assert request.scheme == "http"


def test_static_request_get_host(settings):
    settings.COLTRANE = {"SITE_URL": "http://localhost"}
    request = StaticRequest(path="/")

    assert request.get_host() == "localhost"


def test_static_request_missing_site(settings):
    settings.COLTRANE = {}
    request = StaticRequest(path="/")

    with pytest.raises(AssertionError):
        request.scheme  # noqa: B018


def test_static_request_missing_coltrane(settings):
    delattr(settings, "COLTRANE")
    request = StaticRequest(path="/")

    with pytest.raises(AssertionError):
        request.scheme  # noqa: B018



from copy import deepcopy

from coltrane import DEFAULT_INSTALLED_APPS, _merge_installed_apps


def test_merge_installed_apps_no_installed_apps_in_settings():
    expected = DEFAULT_INSTALLED_APPS
    actual = _merge_installed_apps({}, deepcopy(DEFAULT_INSTALLED_APPS))

    assert actual == expected


def test_merge_installed_apps_installed_apps_in_settings():
    expected = [*deepcopy(DEFAULT_INSTALLED_APPS), "test"]

    django_settings = {"INSTALLED_APPS": ["test"]}
    actual = _merge_installed_apps(django_settings, deepcopy(DEFAULT_INSTALLED_APPS))

    assert actual == expected

    # Make sure that INSTALLED_APPS gets popped out of settings
    assert "INSTALLED_APPS" not in django_settings



from unittest.mock import patch

from coltrane import _get_from_env


@patch("coltrane.getenv", return_value="a,b,c")
def test_get_from_env(getenv):  # noqa: ARG001
    expected = ["a", "b", "c"]
    actual = _get_from_env("TEST_SETTING")

    assert actual == expected



import pytest

from coltrane.retriever import get_content_paths


def test_get_content(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("# test 1")

    contents = list(get_content_paths())

    assert contents
    assert len(contents) == 1
    assert contents[0].name == "test-1.md"


def test_get_content_not_markdown(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("# test 1")
    (tmp_path / "content" / "test-2.test").write_text("# test 2")

    contents = list(get_content_paths())

    assert contents
    assert len(contents) == 1
    assert contents[0].name == "test-1.md"


def test_get_content_with_folder(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test-1.md").write_text("# test 1")

    (tmp_path / "content" / "test-folder").mkdir()
    (tmp_path / "content" / "test-folder" / "test-2.md").write_text("# test 2")

    contents = list(get_content_paths())

    assert contents
    assert len(contents) == 2

    contents.sort(key=lambda c: c.name)
    assert contents[0].name == "test-1.md"
    assert contents[1].name == "test-2.md"


def test_get_content_no_content(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()

    contents = list(get_content_paths())

    assert len(contents) == 0


def test_get_content_missing_content_directory(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    with pytest.raises(FileNotFoundError):
        list(get_content_paths())


def test_get_content_directory_sub_directory_with_md(settings, tmp_path):
    settings.BASE_DIR = tmp_path

    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").mkdir()

    contents = list(get_content_paths())

    assert len(contents) == 0



#!/usr/bin/env python

from pathlib import Path

from django.core.management import execute_from_command_line

from coltrane import initialize

# Pass in this file's parent as the BASE_DIR because app.py lives in a sub-folder
wsgi = initialize(BASE_DIR=Path(__file__).resolve().parent)

if __name__ == "__main__":
    execute_from_command_line()



import pytest

from coltrane.templatetags.coltrane_tags import raise_404


def test_raise_404():
    expected = "django.http.response.Http404"

    with pytest.raises(Exception) as e:
        raise_404()

    actual = e.exconly()

    assert expected in actual


def test_raise_404_with_message():
    expected = "django.http.response.Http404: this is a test"

    with pytest.raises(Exception) as e:
        raise_404("this is a test")

    actual = e.exconly()

    assert expected in actual



import pytest

from coltrane.config.settings import (
    DEFAULT_COLTRANE_SETTINGS,
    get_coltrane_settings,
    get_content_directory,
    get_data_directory,
    get_data_json_5,
    get_description,
    get_extra_file_names,
    get_markdown_renderer,
    get_site_url,
    get_title,
)


def test_get_coltrane_settings_none(settings):
    settings.COLTRANE = {}
    assert get_coltrane_settings() == {}


def test_get_coltrane_settings_default(settings):
    del settings.COLTRANE
    assert get_coltrane_settings() == DEFAULT_COLTRANE_SETTINGS


def test_get_coltrane_settings_invalid_type(settings):
    settings.COLTRANE = "str"

    with pytest.raises(TypeError):
        get_coltrane_settings()


def test_get_site_url(settings):
    settings.COLTRANE = {"SITE_URL": "http://test-site.com"}
    assert get_site_url() == "http://test-site.com"


def test_get_title(settings):
    settings.COLTRANE = {"TITLE": "test-title"}
    assert get_title() == "test-title"


def test_get_description(settings):
    settings.COLTRANE = {"DESCRIPTION": "test description"}
    assert get_description() == "test description"


def test_get_data_directory(settings):
    settings.COLTRANE = {"DATA_DIRECTORY": "test-data-directory"}
    assert get_data_directory() == "test-data-directory"


def test_data_directory_default():
    assert get_data_directory() == "data"


def test_content_directory(settings):
    settings.COLTRANE = {"CONTENT_DIRECTORY": "test-content-directory"}
    assert get_content_directory() == "test-content-directory"


def test_content_directory_default():
    assert get_content_directory() == "content"


def test_get_markdown_renderer(settings):
    assert get_markdown_renderer() == "mistune"

    settings.COLTRANE = {"MARKDOWN_RENDERER": "invalid-markdown-renderer"}

    with pytest.raises(AssertionError):
        get_markdown_renderer()


def test_get_extra_file_names(settings):
    settings.COLTRANE = {"EXTRA_FILE_NAMES": ["robots.txt"]}
    assert get_extra_file_names() == ["robots.txt"]


def test_get_data_json_5(settings):
    assert get_data_json_5() is False

    settings.COLTRANE = {"DATA_JSON5": True}
    assert get_data_json_5() is True



from pathlib import Path

import pytest
from django.template.library import InvalidTemplateLibrary

from coltrane.config.coltrane.configurators.templates import TemplatesConfigurator
from coltrane.config.settings import get_config


def get_templates_configurator(path: str) -> TemplatesConfigurator:
    coltrane = get_config(base_dir=Path(path))

    return TemplatesConfigurator(coltrane)


def test_get_template_tag_module_name():
    expected = "example_standalone.templatetags.custom_tags"

    templates_configurator = get_templates_configurator("../example_standalone")
    actual = templates_configurator._get_template_tag_module_name(
        Path("example_standalone/templatetags/custom_tags.py")
    )

    assert actual == expected


def test_get_template_tag_module_name_invalid_missing_extension():
    templates_configurator = get_templates_configurator("../example_standalone")

    with pytest.raises(InvalidTemplateLibrary):
        templates_configurator._get_template_tag_module_name(Path("../example_standalone"))


def test_get_template_tag_module_name_invalid_wrong_file():
    templates_configurator = get_templates_configurator("../example_standalone")

    with pytest.raises(InvalidTemplateLibrary):
        templates_configurator._get_template_tag_module_name(
            Path("example_standalone/templatetags/custom_tags2.py"),
        )


def test_get_template_tag_module_name_with_period_base_dir():
    expected = "example_standalone.templatetags.custom_tags"

    templates_configurator = get_templates_configurator(".")
    actual = templates_configurator._get_template_tag_module_name(
        Path("example_standalone/templatetags/custom_tags.py"),
    )

    assert actual == expected



from django.conf import settings
from django.urls import include, path, re_path
from django.views.generic.base import RedirectView

from coltrane import views
from coltrane.config.redirects import get_redirects
from coltrane.config.settings import get_extra_file_names
from coltrane.feeds import ContentFeed
from coltrane.sitemaps import ContentSitemap

app_name = "coltrane"

sitemaps = {"content": ContentSitemap}

urlpatterns = []


# Add browser reload URL if not prod
if settings.DEBUG:
    urlpatterns += [
        path("__reload__/", include("django_browser_reload.urls")),
    ]

# Add redirects
for redirect in get_redirects():
    urlpatterns += [
        path(redirect.from_url, RedirectView.as_view(url=redirect.to_url, permanent=redirect.permanent)),
    ]

# Add healthcheck, sitemap, and RSS URLs
urlpatterns += [
    path("healthcheck", views.healthcheck, name="healthcheck"),
    path("sitemap.xml", views.sitemap, name="sitemap"),
    path("rss.xml", ContentFeed()),
]

# Add `django_unicorn` URL if it's installed
if "django_unicorn" in settings.INSTALLED_APPS:
    urlpatterns += [
        path("unicorn/", include("django_unicorn.urls")),
    ]

# Add file names to serve to url routes (e.g. robots.txt)
for file_name in get_extra_file_names():
    urlpatterns += [
        path(file_name, views.file, kwargs={"file_name": file_name}),
    ]

# Add catch-all route for markdown content
urlpatterns += [
    re_path(r"^(?P<slug>(\w|-|\/)*)", views.content, name="content"),
]
